"""
Tests of BSE miscellaneous functions
"""

import pytest
from basis_set_exchange import misc


# yapf: disable
@pytest.mark.parametrize("elements, expected", [([1], "H"),
                                                ([1, 2], "H,He"),
                                                ([1, 10], "H,Ne"),
                                                ([1, 2, 3, 11, 23, 24], "H-Li,Na,V,Cr")])
# yapf: enable
def test_compact_string(elements, expected):
    '''Test compacting a string of elements (and then expanding them)'''
    compacted = misc.compact_elements(elements)
    assert compacted == expected

    expanded = misc.expand_elements(compacted)
    assert expanded == elements

    expanded = misc.expand_elements(compacted, True)
    assert expanded == [str(x) for x in elements]


# yapf: disable
@pytest.mark.parametrize("compact_el, expected",
                               [("H-Li,C-O,Ne", [1, 2, 3, 6, 7, 8, 10]),
                                ("H-N,8,Na-12", [1, 2, 3, 4, 5, 6, 7, 8, 11, 12]),
                                (['C', 'Al-15,S', 17, '18'], [6, 13, 14, 15, 16, 17, 18]),
                                ([], []),
                                ([''], []),
                                (['','',''], []),
                                ('', []),
                                ('C', [6]),
                                (' 6 ', [6]),
                                (6, [6]),
                                (['6,,7', '  10  '], [6,7,10]),
                                (['6----8'], [6,7,8]),
                                ([',,,'], []),
                                ([',6,', ',Ne,,,', ',,11-Si,,'], [6,10,11,12,13,14]),
                                ('6----8', [6,7,8])])
# yapf: enable
def test_expand_elements(compact_el, expected):
    expanded = misc.expand_elements(compact_el)
    assert expanded == expected


# yapf: disable
@pytest.mark.parametrize("compact_el",
                               ['H-Li-C',
                                ['H-Li-C'],
                                '-H',
                                'H-',
                                '1-',
                                '-1',
                                '1,2-,10',
                                '1,-2,10',
                                ['1', '-2', '10']])
# yapf: enable
def test_expand_elements_fail(compact_el):
    with pytest.raises(RuntimeError, match=r'Malformed element string'):
        misc.expand_elements(compact_el)


# yapf: disable
@pytest.mark.parametrize("name,expected",
                              [('6-31g', '6-31g'),
                               ('6-31G', '6-31g'),
                               ('6-31G**', '6-31g_st__st_'),
                               ('6-31++G', '6-31++g'),
                               ('6-31++G**', '6-31++g_st__st_')])
# yapf: enable
def test_basis_name_to_filename(name, expected):
    fname = misc.basis_name_to_filename(name)
    assert fname == expected
    name2 = misc.basis_name_from_filename(fname)
    assert name2 == name.lower()



import pytest


def pytest_addoption(parser):
    parser.addoption("--runslow", action="store_true", default=False, help="Run thorough, but much slower, tests")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--runslow"):
        return

    skip_slow = pytest.mark.skip(reason="Skipping slow test without --runslow option")
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip_slow)


def pytest_ignore_collect(path, config):
    if config.getoption("--runslow"):
        return False
    else:
        return str(path).endswith('_slow.py')



'''
Functions for helping curate BSE basis set data
'''

from .metadata import create_metadata_file
from .add_basis import add_basis, add_from_components
from .compare_report import basis_comparison_report, compare_basis_against_file, compare_basis_files, compare_basis_sets, shells_difference, potentials_difference
from .compare import (compare_electron_shells, electron_shells_are_subset, electron_shells_are_equal, compare_ecp_pots,
                      ecp_pots_are_subset, ecp_pots_are_equal, compare_elements, compare_basis)
from .misc import elements_in_files, component_file_refs
from .graph import view_graph, make_graph_file
from .diff import diff_basis_dict, diff_json_files



'''
Helper functions for writing out basis set references in various formats
'''

from .. import api

_lib_refs = ["pritchard2019a", "feller1996a", "schuchardt2007a"]
_lib_refs_desc = 'If you downloaded data from the basis set\nexchange or used the basis set exchange python library, please cite:\n'


def get_library_citation():
    '''Return a descriptive string and reference data for what users of the library should cite'''

    all_ref_data = api.get_reference_data()
    lib_refs_data = {k: all_ref_data[k] for k in _lib_refs}
    return (_lib_refs_desc, lib_refs_data)



"""
Test for validating the formatting of the json files
"""

import os
import pytest

from basis_set_exchange import api, validator
from .common_testvars import bs_names_vers, all_metadata_paths, all_table_paths, all_element_paths, all_component_paths, data_dir


@pytest.mark.parametrize('file_path', all_metadata_paths)
def test_valid_meta(file_path):
    validator.validate_file('metadata', file_path)


@pytest.mark.parametrize('file_path', all_table_paths)
def test_valid_table(file_path):
    validator.validate_file('table', file_path)


@pytest.mark.parametrize('file_path', all_element_paths)
def test_valid_element(file_path):
    validator.validate_file('element', file_path)


@pytest.mark.parametrize('file_path', all_component_paths)
def test_valid_component(file_path):
    validator.validate_file('component', file_path)


def test_valid_reffile():
    '''
    Test to make sure the references file is valid
    '''
    file_path = os.path.join(data_dir, "REFERENCES.json")
    validator.validate_file('references', file_path)


@pytest.mark.parametrize('bs_name,bs_ver', bs_names_vers)
def test_valid_complete(bs_name, bs_ver):
    '''Test that all basis set data is valid when obtained through get_basis'''
    data = api.get_basis(bs_name, version=bs_ver)
    validator.validate_data('complete', data)



"""
Tests for the BSE main API
"""

import pytest

import basis_set_exchange as bse


@pytest.mark.parametrize('basis_name', ['cc-pv0z', 'madeup_name'])
def test_get_basis_fail_name(basis_name):
    """Test get_basis with bad names """

    with pytest.raises(KeyError, match=r'Basis set.*does not exist'):
        bse.get_basis(basis_name)


@pytest.mark.parametrize('elements', [100, '100', [1, 100], [1, '100']])
def test_get_basis_fail_elements(elements):
    """Test get_basis with bad elements """

    with pytest.raises(KeyError, match=r'Element.*not found in basis'):
        bse.get_basis('cc-pvdz', elements=elements)


@pytest.mark.parametrize('fmt', ['nwchemm', 'nofmt'])
def test_get_basis_fail_fmt(fmt):
    """Test get_basis with bad formats """

    with pytest.raises(RuntimeError, match=r'Unknown basis set format'):
        bse.get_basis('cc-pvdz', fmt=fmt)


@pytest.mark.parametrize('version', ['v1', -1, '-1', 100])
def test_get_basis_fail_version(version):
    """Test get_basis with bad versions """

    with pytest.raises(KeyError, match=r'Version .* does not exist for basis'):
        bse.get_basis('cc-pvdz', version=version)


def test_get_basis_fail_data_dir(tmp_path):
    """Test get_basis with bad data dir """

    tmp_path = str(tmp_path)  # Needed for python 3.5

    with pytest.raises(FileNotFoundError, match=r'JSON file .* does not exist'):
        bse.get_basis('cc-pvdz', data_dir=tmp_path)


def test_get_all_basis_names_fail(tmp_path):
    """Test get_all_basis_names with bad data dir """

    tmp_path = str(tmp_path)  # Needed for python 3.5

    with pytest.raises(FileNotFoundError, match=r'JSON file .* does not exist'):
        bse.get_all_basis_names(tmp_path)


@pytest.mark.parametrize('basis_name', ['cc-pv0z', 'madeup_name'])
def test_lookup_role_fail_name(basis_name):
    """Test get_basis with bad versions """

    with pytest.raises(KeyError, match=r'Basis set.*does not exist'):
        bse.lookup_basis_by_role(basis_name, 'rifit')


@pytest.mark.parametrize('role', ['madeup_name', 'admmfit'])
def test_lookup_role_fail_role(role):
    """Test get_basis with bad role """

    with pytest.raises(RuntimeError, match=r'Role.*(not a valid role|exist for)'):
        bse.lookup_basis_by_role('def2-tzvp', role)


# yapf: disable
@pytest.mark.parametrize('substr,family,role', [['def2', 'ahlrichsx', 'jkfit'],
                                                ['qqqqq', None, 'notarole'],
                                                ['6-31', 'poplexx,', 'admmfit']])
# yapf: enable
def test_filter_fail(substr, family, role):
    """Test filtering basis set (returning zero results)"""

    with pytest.raises(RuntimeError, match=r'(Family|Role).*not a valid (family|role)'):
        md = bse.filter_basis_sets(substr, family, role)



'''
Some helper functions for parsing basis set files
'''

import re
from ..misc import transpose_matrix

floating_re_str = r'[-+]?\d*\.\d*(?:[dDeE][-+]?\d+)?'
floating_re = re.compile(floating_re_str)
floating_only_re = re.compile('^' + floating_re_str + '$')
integer_re_str = r'[-+]?\d+'
integer_re = re.compile(integer_re_str)
integer_only_re = re.compile('^' + integer_re_str + '$')


def _convert_str_int(s):
    '''Optionally convert a string to an integer

    If string s represents an integer, returns an int. Otherwise, returns s unchanged
    '''

    # May throw ValueError if string is not an int
    # May throw TypeError if it is NoneType (some captures may be optional)
    try:
        return int(s)
    except ValueError:
        return s
    except TypeError:
        return s


def is_floating(s):
    '''Tests if a string is a floating point number'''
    return floating_only_re.match(s)


def is_integer(s):
    '''Tests if a string is an integer'''
    return integer_only_re.match(s)


def replace_d(s):
    '''Replaces fortran-style 'D' with 'E'''
    s = s.replace('D', 'E')
    s = s.replace('d', 'e')
    return s


def function_type_from_am(shell_am, base_type, spherical_type):
    if max(shell_am) <= 1:
        return base_type
    else:
        return base_type + '_' + spherical_type


def potential_am_list(max_am):
    '''Creates a canonical list of AM for use with ECP potentials

    The list is [max_am, 0, 1, ..., max_am-1]
    '''

    am_list = list(range(max_am + 1))
    am_list.insert(0, am_list.pop())
    return am_list


def chunk_list(lst, rows, cols):
    '''Turns a list into a matrix of the given dimensions'''

    n_elements = len(lst)
    if n_elements != rows * cols:
        raise RuntimeError("Cannot partition {} elements into a {}x{} matrix".format(n_elements, rows, cols))

    mat = [lst[i:i + cols] for i in range(0, n_elements, cols)]
    assert len(mat) == rows
    return mat


def remove_expected_line(lines, expected='', position=0):
    '''Tests the first element of the list to see if it is an expected string, and removes it

    If line does not match, or lines is empty, an exception is raised
    '''
    if not lines:
        raise RuntimeError("No lines to test for expected line")
    if position >= 0 and len(lines) <= position:
        raise RuntimeError("Not enough lines. Can't test line {} when there are {} lines".format(position, len(lines)))
    if position < 0 and len(lines) < -position:
        raise RuntimeError("Not enough lines. Can't test line {} when there are {} lines".format(position, len(lines)))
    if lines[position] != expected:
        raise RuntimeError("Expected line '{}', but found '{}'".format(expected, lines[0]))

    new_lines = lines[:]
    new_lines.pop(position)
    return new_lines


def create_element_data(bs_data, element_Z, key, key_exist_ok=False, element_exist_ok=True, create=list):
    '''Creates an element and a subkey of the element in bs_data

    Note that bs_data is modified!
    '''

    if element_Z not in bs_data:
        bs_data[element_Z] = {}
    elif not element_exist_ok:
        raise RuntimeError("Element {} already exists in basis data".format(element_Z))

    if key not in bs_data[element_Z]:
        bs_data[element_Z][key] = create()
    elif not key_exist_ok:
        raise RuntimeError("Key {} already exists in basis data for element {}".format(key, element_Z))

    return bs_data[element_Z]


def parse_line_regex(rex, line, description=None, convert_int=True):
    if isinstance(rex, str):
        rex = re.compile(rex)

    r = rex.match(line)
    if not r:
        if description:
            raise RuntimeError("Regex '{}' does not match line: '{}'. Regex is '{}'".format(
                description, line, rex.pattern))
        else:
            raise RuntimeError("Regex '{}' does not match line: '{}'".format(rex.pattern, line))

    g = r.groups()
    if convert_int:
        g = [_convert_str_int(x) for x in g]

    if len(g) == 1:
        return g[0]
    else:
        return g


def partition_lines(lines, condition, before=0, min_after=None, min_blocks=None, max_blocks=None, min_size=1, include_match=True):
    '''Partition a list of lines based on some condition

    Parameters
    ----------
    lines : list
        List of strings representing the lines in the file
    condition : function
        Function or lambda that takes a line as an argument and returns True if
        that line is the start of a section
    before : int
        Number of lines prior to the splitting line (where condition(line) == True)
        to include
`   min_after : int
        Minimum number of lines to include after the match. This number of lines is forced
        even if a match is found again with that set of lines.
    min_blocks : int
        Minimum number of blocks to find. If fewer are found, an exception is thrown
    max_blocks : int
        Maximum number of blocks to find. If more are found, an exception is thrown
    min_size : int
        Minimum size/length of each block. If one is found that is smaller, an exception is thrown

    Returns
    -------
    list of list
        The original list of strings partitioned into blocks
    '''

    # First, just partition into blocks
    all_blocks = []
    cur_block = []

    i = 0
    while i < len(lines):
        line = lines[i]
        if condition(line):
            # This is beginning of a new block. If we have anything,
            # append it to block list
            if cur_block:
                all_blocks.append(cur_block)
                cur_block = []

            if include_match:
                cur_block.append(line)
            if min_after:
                cur_block.extend(lines[i+1:i+1+min_after])
                i += min_after
        else:
            cur_block.append(line)

        i += 1

    # Handle the last block, if needed
    if cur_block:
        all_blocks.append(cur_block)

    # Now handle if the caller specified the 'before' parameter
    if before > 0:
        # Each block after the first steals from the previous
        # If we are going to do this, then:
        #    1.) There must be more than one block
        #    2.) The first block must be only the part to be moved to the second block.
        if len(all_blocks) <= 1:
            raise RuntimeError("Cannot partition lines with before = {}: have {} blocks".format(
                before, len(all_blocks)))

        if len(all_blocks[0]) != before:
            raise RuntimeError("Cannot partition lines with before = {}: first block has {} lines".format(
                before, len(all_blocks[0])))

        # Starting with the second block (index 1), move 'before' lines from the end of the previous block
        # to the beginning of this block
        for idx in range(1, len(all_blocks)):
            all_blocks[idx] = all_blocks[idx - 1][-before:] + all_blocks[idx]
            all_blocks[idx - 1] = all_blocks[idx - 1][:-before]

        # Double check that the first block is now empty (we moved it all)
        first_block = all_blocks.pop(0)
        assert len(first_block) == 0

    # Check to make sure all blocks for min_size
    if min_size > 0:
        for idx, block in enumerate(all_blocks):
            if len(block) < min_size:
                raise RuntimeError("Block {} does not have minimum number of lines ({})".format(idx, min_size))

    if min_blocks and len(all_blocks) < min_blocks:
        raise RuntimeError("Found {} blocks, but need at least {}".format(len(all_blocks), min_blocks))
    if max_blocks and len(all_blocks) > max_blocks:
        raise RuntimeError("Found {} blocks, but need at most {}".format(len(all_blocks), max_blocks))

    return all_blocks


def read_n_floats(lines, n_numbers, convert=False, split=r'\s+'):
    '''Reads in a number of space-separated floating-point numbers

    These numbers may span multiple lines.

    An exception will be thrown if there are remaining numbers on the last line.

    If a non-floating point entry is found, an exception is also thrown.

    If convert is True, then float objects are created from the strings

    Returns the found floating point numbers (as str), and the remaining lines
    '''

    found_numbers = []
    while len(found_numbers) < n_numbers:
        if not lines:
            raise RuntimeError("Wanted {} numbers but ran out of lines after {}".format(n_numbers, len(found_numbers)))
        if not lines[0]:
            raise RuntimeError("Wanted {} numbers but found empty line after {}".format(n_numbers, len(found_numbers)))

        l = replace_d(lines[0])
        s = re.split(split, l.strip())
        found_numbers.extend(s)
        lines = lines[1:]

    if len(found_numbers) != n_numbers:
        raise RuntimeError("Wanted {} numbers, but found {}".format(n_numbers, found_numbers))

    # Make sure these are floating point
    if not all(is_floating(x) for x in found_numbers):
        raise RuntimeError("Non-floating-point value found in numbers: " + ' '.join(found_numbers))

    if convert:
        found_numbers = [float(x) for x in found_numbers]

    return found_numbers, lines


def read_all_floats(lines, convert=False, split=r'\s+'):
    '''Reads in all floats on all lines

    This function takes a block of numbers and splits them all, for all lines in the block.

    If a non-floating point entry is found, an exception is also thrown.

    If convert is True, then float objects are created from the strings
    '''

    found_numbers = []
    for l in lines:
        l = replace_d(l)
        s = re.split(split, l.strip())
        found_numbers.extend(s)

    # Make sure these are floating point
    if not all(is_floating(x) for x in found_numbers):
        raise RuntimeError("Non-floating-point value found in numbers: " + ' '.join(found_numbers))

    if convert:
        found_numbers = [float(x) for x in found_numbers]

    return found_numbers


def read_n_integers(lines, n_ints, convert=False, split=r'\s+'):
    '''Reads in a number of space-separated integers

    These numbers may span multiple lines.

    An exception will be thrown if there are remaining numbers on the last line.

    If a non-integer entry is found, an exception is also thrown.

    If convert is True, then int objects are created from the strings

    Returns the found integers point numbers (as str), and the remaining lines
    '''

    found_numbers = []
    while len(found_numbers) < n_ints:
        s = re.split(split, lines[0].strip())
        found_numbers.extend(s)
        lines = lines[1:]

    if len(found_numbers) != n_ints:
        raise RuntimeError("Wanted {} numbers, but found {}".format(n_ints, found_numbers))

    # Make sure these are integers
    if not all(is_integer(x) for x in found_numbers):
        raise RuntimeError("Non-integer value found in numbers: " + ' '.join(found_numbers))

    if convert:
        found_numbers = [int(x) for x in found_numbers]

    return found_numbers, lines


def parse_fixed_matrix(lines, rows, cols, split=r'\s+'):
    '''Parses a simple matrix of numbers with a predefined number of rows/columns

    This will read in a matrix of the given number of rows and columns, even if the
    rows span multiple lines. There must be a newline at the very end of a row.

    Returns the matrix and the remaining lines
    '''
    mat = []

    for i in range(rows):
        row_data, lines = read_n_floats(lines, cols)
        mat.append(row_data)

    return mat, lines


def parse_matrix(lines, rows=None, cols=None, split=r'\s+'):
    '''Parses a simple matrix of numbers

    The lines parameter must specify a list of strings containing the entire matrix.

    If rows and/or cols is specified, and the found number of rows/cols does not
    match, an exception is raised.
    '''
    mat = []

    for l in lines:
        l = replace_d(l)
        s = re.split(split, l.strip())

        if not all(is_floating(x) for x in s):
            raise RuntimeError("Non-floating-point value found in matrix: " + ' '.join(s))

        mat.append(s)

    # Make sure all rows have the same number of columns
    for x in mat:
        if len(x) == 0:
            raise RuntimeError("Matrix row has zero values")
        if len(x) != len(mat[0]):
            raise RuntimeError("Inconsistent number of columns: {} vs {}".format(len(x), len(mat[0])))

    if len(mat) == 0:
        raise RuntimeError("Empty matrix?")

    # if rows/cols are given, sanity check
    if rows is not None:
        rows = int(rows)

        if len(mat) != rows:
            raise RuntimeError("Inconsistent number of rows: {} vs {}".format(rows, len(mat)))

    # We already checked that all rows have the same number of columns
    # so just check the first row
    if cols is not None and len(mat[0]) != cols:
        raise RuntimeError("Inconsistent number of columns: {} vs {}".format(cols, len(mat[0])))

    return mat


def parse_primitive_matrix(lines, nprim=None, ngen=None, split=r'\s+'):
    '''Parses a matrix/table of exponents and coefficients

    The first column of the matrix contains exponents, and the remaining
    columns contain the coefficients for all general contractions.

    The lines parameter must specify a list of strings containing the entire matrix.

    If nprim and/or ngen are specified, and the found number of primitives/contractions 
    match, an exception is raised.
    '''
    exponents = []
    coefficients = []

    for l in lines:
        l = replace_d(l)
        s = re.split(split, l.strip())

        e = s[0]
        c = s[1:]

        # Make sure all exponents and coefficients are floating point
        if not is_floating(e):
            raise RuntimeError("Non-floating-point value found in exponents: " + e)

        if not all(is_floating(x) for x in c):
            raise RuntimeError("Non-floating-point value found in coefficients: " + ' '.join(c))

        exponents.append(e)
        coefficients.append(c)

    # Make sure all primitives have the same number of coefficients
    for c in coefficients:
        if len(c) == 0:
            raise RuntimeError("Missing contraction coefficients")
        if len(c) != len(coefficients[0]):
            raise RuntimeError("Inconsistent number of coefficients: {} vs {}".format(len(c), len(coefficients[0])))

    coefficients = transpose_matrix(coefficients)

    if len(exponents) == 0:
        raise RuntimeError("No exponents found")
    if len(coefficients) == 0:
        raise RuntimeError("No coefficients found")

    # if nprim and/or ngen are given, sanity check
    if nprim is not None:
        nprim = int(nprim)

        if len(exponents) != nprim:
            raise RuntimeError("Inconsistent number of primitives in exponents: {} vs {}".format(
                nprim, len(exponents)))

        # We already checked that all coefficients have the same number of primitives
        # so just check the first general contraction
        if len(coefficients[0]) != nprim:
            raise RuntimeError("Inconsistent number of primitives in coefficients: {} vs {}".format(
                nprim, len(coefficients[0])))

    if ngen is not None:
        ngen = int(ngen)
        if len(coefficients) != ngen:
            raise RuntimeError("Inconsistent number of general contractions: {} vs {}".format(ngen, len(coefficients)))

    return exponents, coefficients


def parse_ecp_table(lines, order=['r_exp', 'g_exp', 'coeff'], split=r'\s+'):
    ecp_data = {'r_exp': [], 'g_exp': [], 'coeff': []}

    for l in lines:
        l = replace_d(l)
        s = re.split(split, l.strip())

        if len(s) != 3:
            raise RuntimeError("Expected 3 values in ecp table")

        for i, k in enumerate(order):
            ecp_data[k].append(s[i])

    # Note: This function does not handle multiple coefficients
    # So we only have to add another layer to the coefficient list
    ecp_data['coeff'] = [ecp_data['coeff']]

    # Change r exponents to integers
    # But do a check first
    if not all(is_integer(x) for x in ecp_data['r_exp']):
        raise RuntimeError("Non-integer value found in r exponents")

    # Make sure g exponents and coefficients are floating point
    if not all(is_floating(x) for x in ecp_data['g_exp']):
        raise RuntimeError("Non-floating-point value found in g exponents")

    for c in ecp_data['coeff']:
        if not all(is_floating(x) for x in c):
            raise RuntimeError("Non-floating-point value found in coefficients")

    # Now convert r exponents to integers
    ecp_data['r_exp'] = [int(x) for x in ecp_data['r_exp']]

    return ecp_data


def prune_lines(lines, skipchars='', prune_blank=True, strip_end_blanks=True):
    '''Remove comment and blank lines

    Also strips all lines of beginning/ending whitespace.

    Parameters
    ----------
    lines : list of str
        List of lines to prune
    skipchars : str
        Comment characters designating lines to be skipped. Lines starting
        with any character in this string will be removed.
    prune_blank: bool
        Remove blank lines
    strip_end_blank_lines: bool
        Remove starting/ending blank lines (even if prune_blank=False)

    Returns
    -------
    list of str
        Pruned lines
    '''

    lines = [l.strip() for l in lines]

    if skipchars:
        lines = [l for l in lines if len(l) == 0 or l[0] not in skipchars]
    if prune_blank:
        lines = [l for l in lines if l]

    if not lines:
        return lines

    # No need to strip end blank lines if we got rid of all blank lines
    # Note that we can use pop. The lines variable has already been copied
    if strip_end_blanks and not prune_blank:
        while lines and len(lines[0]) == 0:
            lines.pop(0)
        while lines and len(lines[-1]) == 0:
            lines.pop()

    return lines


def remove_block(lines, start_re, end_re):
    '''Removes a block of data from the lines of text
    
       For example, there may be an optional block of options (like in molcas)

       This will only remove a single block

       Parameters
       ----------
       lines : list of str
           Line of text to parse
       start_re : str
           Regex string representing the start of the block (case insensitive)
       end_re : str
           Regex string representing the end of the block (case insensitive)

       Returns
       -------
       list of str, list of str
           The block found (may be empty), the input lines without the block
    '''

    start = re.compile(start_re, flags=re.IGNORECASE)
    end = re.compile(end_re, flags=re.IGNORECASE)

    start_idx = -1
    for idx, l in enumerate(lines):
        if start.match(l):
            if start_idx != -1:
                raise RuntimeError("Multiple blocks starting with '{}' found".format(start))
            start_idx = idx
            break
    else:
        return [], lines  # Block not found

    i = start_idx + 1
    block_lines = []
    while i < len(lines) and not end.match(lines[i]):
        block_lines.append(lines[i])
        i += 1

    # Did we find the end of the block
    if i == len(lines):
        raise RuntimeError("Cannot find end of block. Looking for '{}' to close '{}'".format(end_re, start_re))

    lines = lines[:start_idx] + lines[i + 1:]
    return block_lines, lines



"""
Tests functions for reading in formatted basis sets
"""

import bz2
import os
import pytest

from basis_set_exchange import readers
from .common_testvars import reader_test_data_dir

subdirs = [x for x in os.listdir(reader_test_data_dir)]
subdirs = [os.path.join(reader_test_data_dir, x) for x in subdirs]
subdirs = [x for x in subdirs if os.path.isdir(x)]

test_files = []
for subdir in subdirs:
    subfiles = os.listdir(subdir)
    subfiles = [os.path.join(subdir, x) for x in subfiles if x.endswith('.bz2')]
    subfiles = [os.path.relpath(x, reader_test_data_dir) for x in subfiles]
    test_files.extend(subfiles)

@pytest.mark.parametrize('file_rel_path', test_files)
def test_reader(file_rel_path):
    file_path = os.path.join(reader_test_data_dir, file_rel_path)
    is_bad = ".bad." in os.path.basename(file_path)
    
    if is_bad:
        # Read the first line of the file. This will contain the string
        # that should match the exception
        with bz2.open(file_path, 'rt', encoding='utf-8') as tf:
            match = tf.readline()

        # Get what we want to match as part of the exception
        match = match[8:].strip()

        with pytest.raises(Exception, match=match):
            readers.read_formatted_basis_file(file_path)
    else:
        # Also validate the result (as both component and minimal)
        readers.read_formatted_basis_file(file_path, validate=True, as_component=True)
        readers.read_formatted_basis_file(file_path, validate=True, as_component=False)


def test_reader_equivalent():
    # Test the two dalton formats. These two files are the same basis in the two slightly-different
    # format.
    file1 = os.path.join(reader_test_data_dir, 'dalton', '6-31g.good.1.mol.bz2')
    file2 = os.path.join(reader_test_data_dir, 'dalton', '6-31g.good.2.mol.bz2')
    data1 = readers.read_formatted_basis_file(file1)
    data2 = readers.read_formatted_basis_file(file2)

    assert data1 == data2



import re
from .. import lut
from . import helpers

element_re = re.compile(r'^-?([A-Za-z]{1,3})(?:\s+0)?$')
ecp_shell_start_re = re.compile(r'^([A-Za-z]{1,3})\s+(\d+)\s+(\d+)$')
ecp_am_nelec_re = re.compile(r'^\S+\s+(\d+)\s+(\d+)$')

# This beast of a regex captures all the scaling factors as a single group:
#    '(?:{floating_re_str})+ ' is a non-capturing group of a number of floating point numbers.
#    Then, we capture all of them.
am_line_re = re.compile(r'^([A-Za-z]+)\s+(\d+)((?:\s+{})+)$'.format(helpers.floating_re_str))


def _parse_electron_lines(basis_lines, bs_data):
    '''Parses lines representing all the electron shells for a single element

    Resulting information is stored in bs_data
    '''

    # Last line should be "****"
    last = basis_lines.pop()
    if last != '****':
        raise RuntimeError("Electron shell is missing terminating ****")

    # First line is "{element} 0"
    element_sym = basis_lines[0].split()[0]

    # In the format, if the element symbol starts with a dash, then Gaussian does not crash
    # with an error in case this element does not exist in the molecule input
    # (this is what you need for system basis set libraries).
    element_sym = element_sym.lstrip('-')

    element_Z = lut.element_Z_from_sym(element_sym, as_str=True)
    element_data = helpers.create_element_data(bs_data, element_Z, 'electron_shells')

    # After that come shells. We determine the start of a shell
    # by if the line starts with an angular momentum (a non-numeric character
    shell_blocks = helpers.partition_lines(basis_lines[1:], lambda x: x[0].isalpha())

    for sh_lines in shell_blocks:
        # Shells start with AM nprim scaling
        shell_am, nprim, scaling_factors = helpers.parse_line_regex(am_line_re, sh_lines[0],
                                                                    "Shell AM, nprim, scaling")
        shell_am = lut.amchar_to_int(shell_am, hij=True)
        func_type = helpers.function_type_from_am(shell_am, 'gto', 'spherical')

        # Handle gaussian scaling factors
        # The square of the scaling factor is applied to exponents.
        # Typically they are 1.0, but not always
        scaling_factors = helpers.replace_d(scaling_factors)
        scaling_factors = [float(x) for x in scaling_factors.split()]

        # Remove any scaling factors that are 0.0
        scaling_factors = [x for x in scaling_factors if x != 0.0]

        # We should always have at least one scaling factor
        if len(scaling_factors) == 0:
            raise RuntimeError("No scaling factors given for element {}: Line: {}".format(element_sym, sh_lines[0]))

        # There can be multiple scaling factors, but we don't handle that. It seems to be very rare
        if len(scaling_factors) > 1:
            raise NotImplementedError("Number of scaling factors > 1")

        scaling_factor = float(scaling_factors[0])**2
        has_scaling = scaling_factor != 1.0

        # How many columns of coefficients do we have?
        # Gaussian doesn't support general contractions, so only >1 if
        # you have a fused shell
        ngen = len(shell_am)

        # Now read the exponents and coefficients
        exponents, coefficients = helpers.parse_primitive_matrix(sh_lines[1:], nprim, ngen)

        # If there is a scaling factor, apply it
        # But we keep track of some significant-figure type stuff (as best we can)
        if has_scaling:
            new_exponents = []
            for ex in exponents:
                ex = float(ex) * scaling_factor
                ex = '{:.16E}'.format(ex)

                # Trim useless zeroes
                ex_splt = ex.split('E')
                ex = ex_splt[0].rstrip('0')
                if ex[-1] == '.':  # Stripped all the zeroes...
                    ex += '0'
                ex += 'E' + ex_splt[1]
                new_exponents.append(ex)

            exponents = new_exponents

        shell = {
            'function_type': func_type,
            'region': '',
            'angular_momentum': shell_am,
            'exponents': exponents,
            'coefficients': coefficients
        }

        element_data['electron_shells'].append(shell)


def _parse_ecp_lines(basis_lines, bs_data):
    '''Parses lines representing all the ECP potentials for a single element

    Resulting information is stored in bs_data
    '''

    # First line is "{element} 0", with the zero being optional
    element_sym = basis_lines[0].split()[0]
    element_Z = lut.element_Z_from_sym(element_sym, as_str=True)
    element_data = helpers.create_element_data(bs_data, element_Z, 'ecp_potentials')

    # Second line is information about the ECP
    max_am, ecp_electrons = helpers.parse_line_regex(ecp_am_nelec_re, basis_lines[1], 'ECP max_am, nelec')

    element_data['ecp_electrons'] = ecp_electrons

    # Partition all the potentials
    # Look for lines containing only an integer, but include the line
    # before it (is a comment line)
    ecp_blocks = helpers.partition_lines(basis_lines[2:], helpers.is_integer, before=1)

    for pot_lines in ecp_blocks:
        # first line is comment
        # second line is number of lines

        # Check that the number of lines is consistent
        if not helpers.is_integer(pot_lines[1]):
            raise RuntimeError("Number of lines for potential is not an integer: " + pot_lines[1])

        nlines = int(pot_lines[1])
        if nlines <= 0:
            raise RuntimeError("Number of lines for potential is <= 0")

        if len(pot_lines) != (nlines + 2):
            raise RuntimeError("Number of lines is incorrect. Expected {}, got {}".format(nlines, len(pot_lines) - 2))

        ecp_data = helpers.parse_ecp_table(pot_lines[2:])

        ecp_pot = {
            'angular_momentum': None,
            'ecp_type': 'scalar_ecp',
            'r_exponents': ecp_data['r_exp'],
            'gaussian_exponents': ecp_data['g_exp'],
            'coefficients': ecp_data['coeff']
        }

        element_data['ecp_potentials'].append(ecp_pot)

    # Determine the AM of the potentials
    # Highest AM first, then the rest in order
    all_pot_am = helpers.potential_am_list(max_am)

    # Were there as many potentials as we thought there should be?
    if len(all_pot_am) != len(element_data['ecp_potentials']):
        raise RuntimeError("Found incorrect number of potentials for {}: Expected {}, got {}".format(
            element_sym, len(all_pot_am), len(ecp_potentials)))

    for idx, pot in enumerate(element_data['ecp_potentials']):
        pot['angular_momentum'] = [all_pot_am[idx]]


def read_g94(basis_lines):
    '''Reads G94-formatted file data and converts it to a dictionary with the
       usual BSE fields

       Note that the gaussian format does not store all the fields we
       have, so some fields are left blank
    '''

    # Removes comments
    basis_lines = helpers.prune_lines(basis_lines, '!')

    bs_data = {}

    # Empty file?
    if not basis_lines:
        return bs_data

    # split into element sections (may be electronic or ecp)
    element_sections = helpers.partition_lines(basis_lines, element_re.match, min_size=3)

    for es in element_sections:
        # Try to guess if this is an ecp
        # Each element block starts with the element symbol
        # If the number of lines > 3, and the 4th line is just an integer, then it is an ECP
        if len(es) > 3 and helpers.is_integer(es[3]):
            _parse_ecp_lines(es, bs_data)
        else:
            _parse_electron_lines(es, bs_data)

    return bs_data



'''
Handlers for command line subcommands
'''

from .. import api, bundle, readers, writers, refconverters, convert
from ..misc import compact_elements
from .common import format_columns


def _bse_cli_list_basis_sets(args):
    '''Handles the list-basis-sets subcommand'''
    metadata = api.filter_basis_sets(args.substr, args.family, args.role, args.elements, args.data_dir)

    if args.no_description:
        liststr = [x['display_name'] for x in metadata.values()]
    else:
        liststr = format_columns([(v['display_name'], v['description']) for k, v in metadata.items()])

    return '\n'.join(liststr)


def _bse_cli_list_families(args):
    '''Handles the list-families subcommand'''
    families = api.get_families(args.data_dir)
    return '\n'.join(families)


def _bse_cli_list_writer_formats(args):
    '''Handles the list-writer-formats subcommand'''
    all_formats = writers.get_writer_formats()

    if args.no_description:
        liststr = all_formats.keys()
    else:
        liststr = format_columns(all_formats.items())

    return '\n'.join(sorted(liststr))


def _bse_cli_list_reader_formats(args):
    all_formats = readers.get_reader_formats()

    if args.no_description:
        liststr = all_formats.keys()
    else:
        liststr = format_columns(all_formats.items())

    return '\n'.join(liststr)


def _bse_cli_list_formats(args):
    all_formats = api.get_formats()

    if args.no_description:
        liststr = all_formats.keys()
    else:
        liststr = format_columns(all_formats.items())

    return '\n'.join(liststr)


def _bse_cli_list_ref_formats(args):
    '''Handles the list-ref-formats subcommand'''
    all_refformats = refconverters.get_reference_formats()

    if args.no_description:
        liststr = all_refformats.keys()
    else:
        liststr = format_columns(all_refformats.items())

    return '\n'.join(liststr)


def _bse_cli_list_roles(args):
    '''Handles the list-roles subcommand'''
    all_roles = api.get_roles()

    if args.no_description:
        liststr = all_roles.keys()
    else:
        liststr = format_columns(all_roles.items())

    return '\n'.join(liststr)


def _bse_cli_get_data_dir(args):
    '''Handles the get-data-dir subcommand'''

    return api.get_data_dir()


def _bse_cli_lookup_by_role(args):
    '''Handles the lookup-by-role subcommand'''
    return api.lookup_basis_by_role(args.basis, args.role, args.data_dir)


def _bse_cli_get_basis(args):
    '''Handles the get-basis subcommand'''

    return api.get_basis(
        name=args.basis,
        elements=args.elements,
        version=args.version,
        fmt=args.fmt,
        uncontract_general=args.unc_gen,
        uncontract_spdf=args.unc_spdf,
        uncontract_segmented=args.unc_seg,
        make_general=args.make_gen,
        optimize_general=args.opt_gen,
        data_dir=args.data_dir,
        header=not args.noheader)


def _bse_cli_get_refs(args):
    '''Handles the get-refs subcommand'''
    return api.get_references(
        basis_name=args.basis, elements=args.elements, version=args.version, fmt=args.reffmt, data_dir=args.data_dir)


def _bse_cli_get_info(args):
    '''Handles the get-info subcommand'''

    bs_meta = api.get_metadata(args.data_dir)[args.basis]
    ret = []
    ret.append('-' * 80)
    ret.append(args.basis)
    ret.append('-' * 80)
    ret.append('    Display Name: ' + bs_meta['display_name'])
    ret.append('     Description: ' + bs_meta['description'])
    ret.append('            Role: ' + bs_meta['role'])
    ret.append('          Family: ' + bs_meta['family'])
    ret.append('  Function Types: ' + ','.join(bs_meta['function_types']))
    ret.append('  Latest Version: ' + bs_meta['latest_version'])
    ret.append('')

    aux = bs_meta['auxiliaries']
    if not aux:
        ret.append('Auxiliary Basis Sets: None')
    else:
        ret.append('Auxiliary Basis Sets:')
        ret.extend(format_columns(list(aux.items()), '    '))

    ver = bs_meta['versions']
    ret.append('')
    ret.append('Versions:')

    # Print 4 columns - version, date, elements, revision description
    version_lines = format_columns([(k, v['revdate'], compact_elements(v['elements']), v['revdesc']) for k, v in ver.items()],
                                    '    ')
    ret.extend(version_lines)

    return '\n'.join(ret)


def _bse_cli_get_notes(args):
    '''Handles the get-notes subcommand'''
    return api.get_basis_notes(args.basis, args.data_dir)


def _bse_cli_get_family(args):
    '''Handles the get-family subcommand'''
    return api.get_basis_family(args.basis, args.data_dir)


def _bse_cli_get_versions(args):
    '''Handles the get-versions subcommand'''
    name = args.basis.lower()
    metadata = api.get_metadata(args.data_dir)
    if name not in metadata:
        raise KeyError(
            "Basis set {} does not exist. For a complete list of basis sets, use the 'list-basis-sets' command".format(
                name))

    version_data = {k: v['revdesc'] for k, v in metadata[name]['versions'].items()}

    if args.no_description:
        liststr = version_data.keys()
    else:
        liststr = format_columns(version_data.items())

    return '\n'.join(liststr)


def _bse_cli_get_family_notes(args):
    '''Handles the get-family-notes subcommand'''
    return api.get_family_notes(args.family, args.data_dir)


def _bse_cli_convert_basis(args):
    '''Handles the convert-basis subcommand'''

    # We convert file -> file
    convert.convert_formatted_basis_file(args.input_file, args.output_file, args.in_fmt, args.out_fmt)
    return "Converted {} -> {}".format(args.input_file, args.output_file)


def _bse_cli_create_bundle(args):
    '''Handles the create-bundle subcommand'''
    bundle.create_bundle(args.bundle_file, args.fmt, args.reffmt, args.archive_type, args.data_dir)
    return "Created " + args.bundle_file


def bse_cli_handle_subcmd(args):
    handler_map = {
        'list-formats': _bse_cli_list_writer_formats,
        'list-writer-formats': _bse_cli_list_writer_formats,
        'list-reader-formats': _bse_cli_list_reader_formats,
        'list-ref-formats': _bse_cli_list_ref_formats,
        'list-roles': _bse_cli_list_roles,
        'get-data-dir': _bse_cli_get_data_dir,
        'list-basis-sets': _bse_cli_list_basis_sets,
        'list-families': _bse_cli_list_families,
        'lookup-by-role': _bse_cli_lookup_by_role,
        'get-basis': _bse_cli_get_basis,
        'get-refs': _bse_cli_get_refs,
        'get-info': _bse_cli_get_info,
        'get-notes': _bse_cli_get_notes,
        'get-family': _bse_cli_get_family,
        'get-versions': _bse_cli_get_versions,
        'get-family-notes': _bse_cli_get_family_notes,
        'convert-basis': _bse_cli_convert_basis,
        'create-bundle': _bse_cli_create_bundle
    }

    return handler_map[args.subcmd](args)



'''
Functions for creating archives of all basis sets
'''

import os
import copy
import zipfile
import tarfile
import io
import datetime
from . import api, writers, refconverters, misc

_readme_str = '''Basis set exchange: Basis set bundle
==========================================

Basis Set Exchange: {bsever}
Created {timestamp}
Format: {fmt}
Reference format: {reffmt}

This directory contains all the basis sets in the library
in {fmt} format.

Filenames of the basis sets are in the format
{{name}}.{{version}}.{{extension}} where the version represents
the version of the basis set.

Filenames of the references are similar, except they
contain .ref before the final extension.

Basis set notes have a .notes extension, and family
notes have a .family_notes extension.

-------------------------------------------------
https://www.basissetexchange.org
https://github.com/MolSSI-BSE/basis_set_exchange
bse@molssi.org
-------------------------------------------------
'''


def _create_readme(fmt, reffmt):
    '''
    Creates the readme file for the bundle

    Returns a str representing the readme file
    '''

    now = datetime.datetime.utcnow()
    timestamp = now.strftime('%Y-%m-%d %H:%M:%S UTC')

    # yapf: disable
    outstr = _readme_str.format(timestamp=timestamp,
                                bsever=api.version(),
                                fmt=fmt, reffmt=reffmt)
    # yapf: enable

    return outstr


def _basis_data_iter(fmt, reffmt, data_dir):
    '''Iterate over all basis set names, and return a tuple of
       (name, data) where data is the basis set in the given format
    '''
    md = api.get_metadata(data_dir)
    for bs, bs_md in md.items():
        versions = bs_md['versions'].keys()

        data = {}
        for v in versions:
            bsdata = api.get_basis(bs, fmt=fmt, version=v, data_dir=data_dir)
            refdata = api.get_references(bs, fmt=reffmt, version=v, data_dir=data_dir)
            data[v] = (bsdata, refdata)

        notes = api.get_basis_notes(bs, data_dir)
        yield (bs, data, notes)


def _add_to_tbz(tfile, filename, data_str):
    '''
    Adds string data to a tarfile
    '''

    # Create a bytesio object for adding to a tarfile
    # https://stackoverflow.com/a/52724508
    encoded_data = data_str.encode('utf-8')
    ti = tarfile.TarInfo(name=filename)
    ti.size = len(encoded_data)
    tfile.addfile(tarinfo=ti, fileobj=io.BytesIO(encoded_data))


def _add_to_zip(zfile, filename, data_str):
    '''
    Adds string data to a zipfile
    '''
    zfile.writestr(filename, data_str)


def _bundle_tbz(outfile, fmt, reffmt, data_dir):
    with tarfile.open(outfile, 'w:bz2') as tf:
        _bundle_generic(tf, _add_to_tbz, fmt, reffmt, data_dir)


def _bundle_zip(outfile, fmt, reffmt, data_dir):
    with zipfile.ZipFile(outfile, 'w', zipfile.ZIP_DEFLATED) as zf:
        _bundle_generic(zf, _add_to_zip, fmt, reffmt, data_dir)


def _bundle_generic(bfile, addhelper, fmt, reffmt, data_dir):
    '''
    Loop over all basis sets and add data to an archive

    Parameters
    ----------
    bfile : object
        An object that gets passed through to the addhelper function
    addhelper : function
        A function that takes bfile and adds data to the bfile
    fmt : str
        Format of the basis set to create
    reffmt : str
        Format to use for the references
    data_dir : str
        Data directory with all the basis set information.

    Returns
    -------
    None
    '''

    ext = writers.get_format_extension(fmt)
    refext = refconverters.get_format_extension(reffmt)
    subdir = 'basis_set_bundle-' + fmt + '-' + reffmt

    readme_path = os.path.join(subdir, 'README.txt')
    addhelper(bfile, readme_path, _create_readme(fmt, reffmt))

    for name, data, notes in _basis_data_iter(fmt, reffmt, data_dir):
        for ver, verdata in data.items():
            filename = misc.basis_name_to_filename(name)
            basis_filepath = os.path.join(subdir, '{}.{}{}'.format(filename, ver, ext))
            ref_filename = os.path.join(subdir, '{}.{}.ref{}'.format(filename, ver, refext))

            bsdata, refdata = verdata
            addhelper(bfile, basis_filepath, bsdata)
            addhelper(bfile, ref_filename, refdata)

        if notes:
            notes_filename = os.path.join(subdir, filename + '.notes')
            addhelper(bfile, notes_filename, notes)

    for fam in api.get_families(data_dir):
        fam_notes = api.get_family_notes(fam, data_dir)

        if fam_notes:
            fam_notes_filename = os.path.join(subdir, fam + '.family_notes')
            addhelper(bfile, fam_notes_filename, fam_notes)


_bundle_types = {
    'zip': {
        'display': 'ZIP file',
        'handler': _bundle_zip,
        'extension': '.zip'
    },
    'tbz': {
        'display': 'Tar + bzip2',
        'handler': _bundle_tbz,
        'extension': '.tar.bz2'
    }
}


def create_bundle(outfile, fmt, reffmt, archive_type=None, data_dir=None):
    '''
    Create a single archive file containing all basis
    sets in a given format

    Parameters
    ----------
    outfile : str
        Path to the file to create. Existing files will be overwritten
    fmt : str
        Format of the basis set to archive (nwchem, turbomole, ...)
    reffmt : str
        Format of the basis set references to archive (nwchem, turbomole, ...)
    archive_type : str
        Type of archive to create. Can be 'zip' or 'tbz'. Default is
        None, which will autodetect based on the outfile name
    data_dir : str
        Data directory with all the basis set information. By default,
        it is in the 'data' subdirectory of this project.

    Returns
    -------
    None
    '''

    if archive_type is None:
        outfile_lower = outfile.lower()

        for k, v in _bundle_types.items():
            if outfile_lower.endswith(v['extension']):
                archive_type = k
                break
        else:
            raise RuntimeError("Cannot autodetect archive type from file name: {}".format(os.path.basename(outfile)))

    else:
        archive_type = archive_type.lower()
        if archive_type not in _bundle_types:
            raise RuntimeError("Archive type '{}' is not valid.")

    _bundle_types[archive_type]['handler'](outfile, fmt, reffmt, data_dir)


def get_archive_types():
    '''
    Return information related to the types of archives available
    '''
    ret = copy.deepcopy(_bundle_types)
    for k, v in ret.items():
        v.pop('handler')
    return ret



'''
Conversion of basis sets to ORCA
'''

from .. import lut, printing
from .gamess_us import write_gamess_us_common


def write_orca_ecp_basis(basis, ecp_elements):
    s = ''
    for z in ecp_elements:
        s += '\n\n'
        data = basis['elements'][z]
        sym = lut.element_sym_from_Z(z).upper()
        max_ecp_am = max([x['angular_momentum'][0] for x in data['ecp_potentials']])
        max_ecp_amchar = lut.amint_to_char([max_ecp_am], hij=True)

        # Sort lowest->highest
        ecp_list = sorted(data['ecp_potentials'], key=lambda x: x['angular_momentum'])

        # Could probably be basis['names'][0]-ECP, but seems like special characters
        # would cause problems
        ecp_name = 'NewECP'
        s += '{} {}\n'.format(ecp_name, sym)
        s += '  N_core {}\n'.format(data['ecp_electrons'])
        s += '  lmax {}\n'.format(max_ecp_amchar)

        for pot in ecp_list:
            rexponents = pot['r_exponents']
            gexponents = pot['gaussian_exponents']
            coefficients = pot['coefficients']
            nprim = len(rexponents)

            am = pot['angular_momentum']
            amchar = lut.amint_to_char(am, hij=False)

            # Title line
            s += '  {} {}\n'.format(amchar, nprim)

            # Include an index column
            idx_column = list(range(1, nprim + 1))
            point_places = [4, 12, 27, 36]
            s += printing.write_matrix([idx_column, gexponents, *coefficients, rexponents], point_places)

        s += "end"
    return s


def write_orca(basis):
    '''Converts a basis set to ORCA
    '''

    return write_gamess_us_common(basis, write_orca_ecp_basis)



'''
Comparison of basis data against authoritative sources
'''

from ..api import get_basis
from ..misc import compact_elements
from ..sort import sort_shells, sort_potentials
from .. import manip
from ..readers import read_formatted_basis_file
from .compare import _reldiff


def _print_list(lst):
    '''
    Prints a list from a comparison
    '''

    a = compact_elements(lst)
    if a is not None:
        return a
    else:
        return ""


def shells_difference(s1, s2):
    """
    Computes and prints the differences between two lists of shells

    If the shells contain a different number primitives,
    or the lists are of different length, inf is returned.
    Otherwise, the maximum relative difference is returned.
    """

    max_rdiff = 0.0
    nsh = len(s1)
    if len(s2) != nsh:
        print("Different number of shells: {} vs {}".format(len(s1), len(s2)))
        return float('inf')

    shells1 = sort_shells(s1)
    shells2 = sort_shells(s2)

    for n in range(nsh):
        sh1 = shells1[n]
        sh2 = shells2[n]

        if sh1['angular_momentum'] != sh2['angular_momentum']:
            print("Different angular momentum for shell {}".format(n))
            return float('inf')

        nprim = len(sh1['exponents'])
        if len(sh2['exponents']) != nprim:
            print("Different number of primitives for shell {}".format(n))
            return float('inf')

        ngen = len(sh1['coefficients'])
        if len(sh2['coefficients']) != ngen:
            print("Different number of general contractions for shell {}".format(n))
            return float('inf')

        for p in range(nprim):
            e1 = sh1['exponents'][p]
            e2 = sh2['exponents'][p]
            r = _reldiff(e1, e2)
            if r > 0.0:
                print("   Exponent {:3}: {:20} {:20} -> {:16.8e}".format(p, e1, e2, r))
            max_rdiff = max(max_rdiff, r)

            for g in range(ngen):
                c1 = sh1['coefficients'][g][p]
                c2 = sh2['coefficients'][g][p]
                r = _reldiff(c1, c2)
                if r > 0.0:
                    print("Coefficient {:3}: {:20} {:20} -> {:16.8e}".format(p, c1, c2, r))
                max_rdiff = max(max_rdiff, r)

    print()
    print("Max relative difference for these shells: {}".format(max_rdiff))
    return max_rdiff


def potentials_difference(p1, p2):
    """
    Computes and prints the differences between two lists of potentials

    If the shells contain a different number primitives,
    or the lists are of different length, inf is returned.
    Otherwise, the maximum relative difference is returned.
    """

    max_rdiff = 0.0
    np = len(p1)
    if len(p2) != np:
        print("Different number of potentials")
        return float('inf')

    pots1 = sort_potentials(p1)
    pots2 = sort_potentials(p2)

    for n in range(np):
        pot1 = pots1[n]
        pot2 = pots2[n]

        if pot1['angular_momentum'] != pot2['angular_momentum']:
            print("Different angular momentum for potential {}".format(n))
            return float('inf')

        nprim = len(pot1['gaussian_exponents'])
        if len(pot2['gaussian_exponents']) != nprim:
            print("Different number of primitives for potential {}".format(n))
            return float('inf')

        ngen = len(pot1['coefficients'])
        if len(pot2['coefficients']) != ngen:
            print("Different number of general contractions for potential {}".format(n))
            return float('inf')

        for p in range(nprim):
            e1 = pot1['gaussian_exponents'][p]
            e2 = pot2['gaussian_exponents'][p]
            r = _reldiff(e1, e2)
            if r > 0.0:
                print("   Gaussian Exponent {:3}: {:20} {:20} -> {:16.8e}".format(p, e1, e2, r))
            max_rdiff = max(max_rdiff, r)

            e1 = pot1['r_exponents'][p]
            e2 = pot2['r_exponents'][p]
            r = _reldiff(e1, e2)
            if r > 0.0:
                print("          R Exponent {:3}: {:20} {:20} -> {:16.8e}".format(p, e1, e2, r))
            max_rdiff = max(max_rdiff, r)

            for g in range(ngen):
                c1 = pot1['coefficients'][g][p]
                c2 = pot2['coefficients'][g][p]
                r = _reldiff(c1, c2)
                if r > 0.0:
                    print("         Coefficient {:3}: {:20} {:20} -> {:16.8e}".format(p, c1, c2, r))
                max_rdiff = max(max_rdiff, r)

    print()
    print("Max relative difference for these potentials: {}".format(max_rdiff))
    return max_rdiff


def basis_comparison_report(bs1, bs2, uncontract_general=False):
    '''
    Compares two basis set dictionaries and prints a report about their differences
    '''

    all_bs1 = list(bs1['elements'].keys())

    if uncontract_general:
        bs1 = manip.uncontract_general(bs1)
        bs2 = manip.uncontract_general(bs2)

    not_in_bs1 = []  # Found in bs2, not in bs1
    not_in_bs2 = all_bs1.copy()  # Found in bs1, not in bs2
    no_diff = []  # Elements for which there is no difference
    some_diff = []  # Elements that are different
    big_diff = []  # Elements that are substantially different

    for k, v in bs2['elements'].items():
        if k not in all_bs1:
            not_in_bs1.append(k)
            continue

        print()
        print("-------------------------------------")
        print(" Element ", k)
        bs1_el = bs1['elements'][k]

        max_rdiff_el = 0.0
        max_rdiff_ecp = 0.0

        # Check to make sure that neither or both have ecp/electron shells
        if 'electron_shells' in v and 'electron_shells' not in bs1_el:
            print("bs2 has electron_shells, but bs1 does not")
            max_rdiff_el = float('inf')
        if 'electron_shells' in bs1_el and 'electron_shells' not in v:
            print("bs1 has electron_shells, but bs2 does not")
            max_rdiff_el = float('inf')
        if 'ecp_potentials' in v and 'ecp_potentials' not in bs1_el:
            print("bs2 has ecp_potentials, but bs1 does not")
            max_rdiff_ecp = float('inf')
        if 'ecp_potentials' in bs1_el and 'ecp_potentials' not in v:
            print("bs1 has ecp_potentials, but bs2 does not")
            max_rdiff_ecp = float('inf')

        if 'electron_shells' in v and 'electron_shells' in bs1_el:
            max_rdiff_el = max(max_rdiff_el, shells_difference(v['electron_shells'], bs1_el['electron_shells']))
        if 'ecp_potentials' in v and 'ecp_potentials' in bs1_el:
            nel1 = v['ecp_electrons']
            nel2 = bs1_el['ecp_electrons']
            if int(nel1) != int(nel2):
                print('Different number of electrons replaced by ECP ({} vs {})'.format(nel1, nel2))
                max_rdiff_ecp = float('inf')
            else:
                max_rdiff_ecp = max(max_rdiff_ecp, potentials_difference(v['ecp_potentials'],
                                                                         bs1_el['ecp_potentials']))

        max_rdiff = max(max_rdiff_el, max_rdiff_ecp)

        # Handle some differences
        if max_rdiff == float('inf'):
            big_diff.append(k)
        elif max_rdiff == 0.0:
            no_diff.append(k)
        else:
            some_diff.append(k)

        not_in_bs2.remove(k)

    print()
    print("     Not in bs1: ", _print_list(not_in_bs1))
    print("     Not in bs2: ", _print_list(not_in_bs2))
    print("  No difference: ", _print_list(no_diff))
    print("Some difference: ", _print_list(some_diff))
    print(" BIG difference: ", _print_list(big_diff))
    print()

    return (not not_in_bs1 and not not_in_bs2 and not some_diff and not big_diff)


def compare_basis_against_file(basis_name,
                               src_filepath,
                               file_type=None,
                               version=None,
                               uncontract_general=False,
                               data_dir=None):
    '''Compare a basis set in the BSE against a reference file'''

    src_data = read_formatted_basis_file(src_filepath, file_type)
    bse_data = get_basis(basis_name, version=version, data_dir=data_dir)
    return basis_comparison_report(src_data, bse_data, uncontract_general=uncontract_general)


def compare_basis_files(file_path_1, file_path_2, file_type_1=None, file_type_2=None, uncontract_general=False):
    '''Compare two files containing formatted basis sets'''

    bs1 = read_formatted_basis_file(file_path_1, file_type_1)
    bs2 = read_formatted_basis_file(file_path_2, file_type_2)
    return basis_comparison_report(bs1, bs2, uncontract_general)


def compare_basis_sets(basis_name_1,
                       basis_name_2,
                       version_1=None,
                       version_2=None,
                       uncontract_general=False,
                       data_dir_1=None,
                       data_dir_2=None):
    '''Compare two files containing formatted basis sets'''

    bs1 = get_basis(basis_name_1, version=version_1, data_dir=data_dir_1)
    bs2 = get_basis(basis_name_2, version=version_2, data_dir=data_dir_2)
    return basis_comparison_report(bs1, bs2, uncontract_general)



'''
Converts basis set data to a specified output format
'''

import textwrap
import json

from .. import sort, misc
from .common import get_library_citation
from .bib import write_bib
from .ris import write_ris
from .endnote import write_endnote

# For plain text
from ..references import reference_text

_converter_map = {
    'txt': {
        'display': 'Plain Text',
        'extension': '.txt',
        'comment': '',
        'function': reference_text
    },
    'bib': {
        'display': 'BibTeX',
        'extension': '.bib',
        'comment': '%',
        'function': write_bib
    },
    'ris': {
        'display': 'RIS',
        'extension': '.RIS',
        'comment': '#',
        'function': write_ris
    },
    'endnote': {
        'display': 'EndNote',
        'extension': '.enw',
        'comment': '#',
        'function': write_endnote
    },
    'json': {
        'display': 'JSON',
        'extension': '.json',
        'comment': '',
        'function': None  # Handled separately
    }
}


def convert_references(ref_data, fmt):
    '''
    Returns the basis set references as a string representing
    the data in the specified output format
    '''

    # Make fmt case insensitive
    fmt = fmt.lower()
    if fmt not in _converter_map:
        raise RuntimeError('Unknown reference format "{}"'.format(fmt))

    # Shortcut for JSON
    if fmt == 'json':
        return json.dumps(ref_data, indent=4, ensure_ascii=False)

    # Sort the data for all references
    for elref in ref_data:
        for rinfo in elref['reference_info']:
            rdata = rinfo['reference_data']
            rinfo['reference_data'] = [(k, sort.sort_single_reference(v)) for k, v in rdata]

    # This function is used to convert a single ref
    single_ref_func = _converter_map[fmt]['function']

    # Comment style
    comment = _converter_map[fmt]['comment']
    comment_line = comment * 80 + '\n'

    # Actually do the conversion
    ref_str = ''

    # First, convert the library citations (for citing the BSE)
    lib_citation_desc, lib_citations = get_library_citation()

    ref_str += comment_line
    ref_str += textwrap.indent(lib_citation_desc, comment + ' ')
    ref_str += comment_line

    for k, r in lib_citations.items():
        ref_str += single_ref_func(k, r) + '\n\n'

    ref_str += comment_line
    ref_str += comment + " References for the basis set\n"
    ref_str += comment_line

    # First, write out the element, description -> key mapping
    # Also make a dict of unique reference to output
    unique_refs = {}

    for ref in ref_data:
        ref_str += comment + ' {}\n'.format(misc.compact_elements(ref['elements']))

        for ri in ref['reference_info']:
            ref_str += comment + '     {}\n'.format(ri['reference_description'])

            single_ref_data = ri['reference_data']

            if len(single_ref_data) == 0:
                ref_str += comment + '         (...no reference...)\n' + comment + '\n'
            else:
                rkeys = [x[0] for x in ri['reference_data']]
                ref_str += comment + '         {}\n{}\n'.format(' '.join(rkeys), comment)

            for k, r in single_ref_data:
                unique_refs[k] = r

    ref_str += '\n\n'

    # Go through them sorted alphabetically by key
    for k, r in sorted(unique_refs.items(), key=lambda x: x[0]):
        ref_str += '{}\n\n'.format(single_ref_func(k, r))

    return ref_str


def get_reference_formats():
    '''Return information about the reference/citation formats available

    The returned data is a map of format to display name. The format
    can be passed as the fmt argument to :func:`get_references`
    '''

    return {k: v['display'] for k, v in _converter_map.items()}


def get_format_extension(fmt):
    '''
    Returns the recommended extension for a given format
    '''

    if fmt is None:
        return 'dict'

    fmt = fmt.lower()
    if fmt not in _converter_map:
        raise RuntimeError('Unknown basis set format "{}"'.format(fmt))

    return _converter_map[fmt]['extension']



"""
Tests for creating bundles/archives of formatted data
"""

import pytest

from .test_bundle import _run_test_bundles, _bundle_exts

# yapf: disable
@pytest.mark.slow
@pytest.mark.parametrize('ext', _bundle_exts)
@pytest.mark.parametrize('fmt, reffmt', [('nwchem', 'bib'),
                                         ('psi4', 'txt'),
                                         ('json', 'json')])
# yapf: enable
def test_bundles_slow(tmp_path, fmt, reffmt, ext):
    _run_test_bundles(tmp_path, fmt, reffmt, ext, None)



'''
Test for duplicate data in a basis set
'''

import pytest
import basis_set_exchange as bse
from .common_testvars import bs_names_vers, true_false
from .test_duplicate import _test_duplicates


@pytest.mark.slow
@pytest.mark.parametrize('bs_name,bs_ver', bs_names_vers)
@pytest.mark.parametrize('unc_gen', true_false)
@pytest.mark.parametrize('unc_seg', true_false)
@pytest.mark.parametrize('unc_spdf', true_false)
@pytest.mark.parametrize('make_gen', true_false)
@pytest.mark.parametrize('opt_gen', true_false)
def test_duplicate_slow(bs_name, bs_ver, unc_gen, unc_seg, unc_spdf, opt_gen, make_gen):
    '''
    Test for any duplicate data in a basis set
    '''

    bs_dict = bse.get_basis(bs_name, version=bs_ver,
                            uncontract_general=unc_gen,
                            uncontract_segmented=unc_seg,
                            uncontract_spdf=unc_spdf,
                            make_general=make_gen,
                            optimize_general=opt_gen)

    _test_duplicates(bs_dict, False)



'''
Conversion of basis sets to deMon2K format
'''

from .. import lut, manip, sort, misc, printing


def write_demon2k(basis):
    '''Converts a basis set to deMon2K format
    '''

    if 'gto_spherical' in basis['function_types']:
        s = '# This basis set uses spherical components\n\n'
    else:
        s = '# This basis set uses cartesian components\n\n'

    basis = manip.uncontract_spdf(basis, 0, True)
    basis = manip.uncontract_general(basis, False)
    basis = sort.sort_basis(basis, False)

    # Elements for which we have electron basis
    electron_elements = [k for k, v in basis['elements'].items() if 'electron_shells' in v]

    # Elements for which we have ECP
    ecp_elements = [k for k, v in basis['elements'].items() if 'ecp_potentials' in v]

    # Electron Basis
    if electron_elements:
        for z in electron_elements:
            data = basis['elements'][z]
            sym = lut.element_sym_from_Z(z, True)
            elname = lut.element_name_from_Z(z).upper()
            cont_string = misc.contraction_string(data)

            # Need the start of electron shells if there are ECPs
            ecp_electrons = data.get('ecp_electrons', 0)
            shells_start = lut.electron_shells_start(ecp_electrons)
            shells_start = list(shells_start)

            s += 'O-{} {} ({})\n'.format(elname, sym.upper(), basis['name'])
            s += '# {}\n'.format(cont_string)

            nshells = len(data['electron_shells'])
            s += '    {}\n'.format(nshells)

            for shell in data['electron_shells']:
                exponents = shell['exponents']
                coefficients = shell['coefficients']
                ncol = len(coefficients) + 1
                nprim = len(exponents)

                # We removed spdf already
                assert len(shell['angular_momentum']) == 1
                am = shell['angular_momentum'][0]

                # shells_start has starting principal quantum numbers for all AM
                pqn = shells_start[am]
                shells_start[am] += 1
                s += '    {}    {}    {}\n'.format(pqn, am, nprim)

                point_places = [8 * i + 15 * (i - 1) for i in range(1, ncol + 1)]
                s += printing.write_matrix([exponents, *coefficients], point_places, convert_exp=False)

    # Write out ECP
    if ecp_elements:
        s += '\n\nECP\n'
        for z in ecp_elements:
            data = basis['elements'][z]
            sym = lut.element_sym_from_Z(z, normalize=True)
            max_ecp_am = max([x['angular_momentum'][0] for x in data['ecp_potentials']])

            # Sort lowest->highest, then put the highest at the beginning
            ecp_list = sorted(data['ecp_potentials'], key=lambda x: x['angular_momentum'])
            ecp_list.insert(0, ecp_list.pop())

            s += '{} nelec {}\n'.format(sym, data['ecp_electrons'])

            for pot in ecp_list:
                rexponents = pot['r_exponents']
                gexponents = pot['gaussian_exponents']
                coefficients = pot['coefficients']

                am = pot['angular_momentum']
                amchar = lut.amint_to_char(am).upper()

                if am[0] == max_ecp_am:
                    s += '{} ul\n'.format(sym)
                else:
                    s += '{} {}\n'.format(sym, amchar)

                point_places = [0, 9, 32]
                s += printing.write_matrix([rexponents, gexponents, *coefficients], point_places, convert_exp=False)

        s += 'END\n'

    return s



"""
Functions related to validating JSON files (including against schema)
"""

import os
import jsonschema
import datetime

from . import fileio, misc

_my_dir = os.path.dirname(os.path.abspath(__file__))
_default_schema_dir = os.path.join(_my_dir, 'schema')


def _list_has_duplicates(lst):
    '''Check if a list has only unique elements
       Returns a list of duplicated elements'''

    dupe = []
    for x in lst:
        if lst.count(x) != 1:
            dupe.append(x)
    return dupe


def _validate_extra_references(bs_data):
    '''Extra checks for references files'''
    pass


def _validate_extra_metadata(bs_data):
    '''Extra checks for metadata files'''

    # Check that family is lowercase
    fam = bs_data['family']
    if not fam.islower():
        raise RuntimeError("Family '{}' is not lowercase".format(fam))


def _validate_electron_shells(shells, element_z):
    '''Validate a list of electron shells'''

    for idx, s in enumerate(shells):
        nprim = len(s['exponents'])
        if nprim <= 0:
            raise RuntimeError("Element {} Shell {}: Invalid number of primitives: {}".format(element_z, idx, nprim))

        # If max(am) > 1, gto types must specify spherical or cartesian
        if max(s['angular_momentum']) > 1:
            if s['function_type'] not in ['gto_spherical', 'gto_cartesian']:
                raise RuntimeError("Element {} Shell {}: Shell with max am > p, but spherical/cartesian not specified".format(element_z, idx))
        else:
            # If max(am) < 2, spherical/cartesian not allowed
            if 'spherical' in s['function_type'] or 'cartesian' in s['function_type']:
                raise RuntimeError("Element {} Shell {}: AM = {} marked as spherical or cartesian: {}".format(element_z, idx, str(s['angular_momentum']), s['function_type']))
             

        # Duplicate exponents (when converted to float)?
        exponents_f = [float(x) for x in s['exponents']]
        dupe_ex = _list_has_duplicates(exponents_f)
        if dupe_ex:
            raise RuntimeError("Element {} Shell {}: Has duplicate exponents: {}".format(element_z, idx, dupe_ex))

        for g in s['coefficients']:
            if nprim != len(g):
                raise RuntimeError(
                    "Element {} Shell {}: Number of coefficients doesn't match number of primitives ({} vs {})".format(
                        element_z, idx, len(g), nprim))

            # Column of zero coefficients?
            coefficients_f = [float(x) for x in g]
            if all(x == 0.0 for x in coefficients_f):
                raise RuntimeError("Element {} Shell {}: Has column of coefficients with all = 0.0".format(
                    element_z, idx))

        # Duplicate columns of coefficients?
        # Only test this is not a fused shell (which can have duplicates)
        all_coefficients_f = [[float(x) for x in g] for g in s['coefficients']]
        if len(s['angular_momentum']) == 1:
            dupe_coef_col = _list_has_duplicates(all_coefficients_f)
            if dupe_coef_col:
                raise RuntimeError("Element {} Shell {}: Duplicate columns of coefficients: ".format(
                    element_z, idx, dupe_coef_col))

        # Does a primitive have all zeroes in the coefficients?
        coeff_t = misc.transpose_matrix(all_coefficients_f)
        for pidx, row in enumerate(coeff_t):
            if all(x == 0.0 for x in row):
                raise RuntimeError("Element {} Shell {} Primitive {}: Primitive is unused (all coeffs = 0.0)".format(
                    element_z, idx, pidx))

        # If more than one AM is given, that should be the number of
        # general contractions
        nam = len(s['angular_momentum'])
        if nam > 1:
            ngen = len(s['coefficients'])
            if ngen != nam:
                raise RuntimeError(
                    "Element {} Shell {}: Number of general contractions doesn't match combined AM ({} vs {})".format(
                        element_z, idx, ngen, nam))


def _validate_ecp_potentials(potentials, ecp_electrons, element_z):
    # Check for duplicate AM and 'fused' AM
    all_am = [x['angular_momentum'] for x in potentials]
    for am in all_am:
        if len(am) > 1:
            raise RuntimeError("Element {} ECP: Fused AM in potentials (not supported)".format(element_Z))

    all_am = [x[0] for x in all_am]
    dupe_am = _list_has_duplicates(all_am)
    if dupe_am:
        raise RuntimeError("Element {} ECP: Duplicated angular momentum: {}".format(dupe_am))

    # Need this for later
    max_am = max(pot['angular_momentum'] for pot in potentials)

    for idx, pot in enumerate(potentials):
        nexp = len(pot['r_exponents'])
        if len(pot['gaussian_exponents']) != nexp:
            raise RuntimeError("Element {} ECP Potential {}: len(r_exponents) != len(gaussian_exponents)".format(
                element_z, idx))

        for g in pot['coefficients']:
            if nexp != len(g):
                raise RuntimeError(
                    "Element {} ECP Potential {}: Number of coefficients doesn't match number of exponents ({} vs {})".
                    format(element_z, idx, len(g), nexp))

            # Column of zero coefficients?
            # Sometimes there is a potential with one exponent and a zero coefficient, but that should be the highest AM
            if nexp > 1 or pot['angular_momentum'] != max_am:
                coefficients_f = [float(x) for x in g]
                if all(x == 0.0 for x in coefficients_f):
                    raise RuntimeError("Element {} ECP Potential {}: Has column of coefficients with all = 0.0".format(
                        element_z, idx))

        # Duplicated columns of coefficients?
        all_coefficients_f = [[float(x) for x in g] for g in pot['coefficients']]
        dupe_coef_col = _list_has_duplicates(all_coefficients_f)
        if dupe_coef_col:
            raise RuntimeError("Element {} ECP Potential {}: Duplicate columns of coefficients: ".format(
                element_z, idx, dupe_coef_col))

        # Check for rows with 0.0 in the coefficients, except for max_am with nexp == 1
        if nexp > 1 or pot['angular_momentum'] != max_am:
            coeff_t = misc.transpose_matrix(all_coefficients_f)
            for pidx, row in enumerate(coeff_t):
                if all(x == 0.0 for x in row):
                    raise RuntimeError(
                        "Element {} Shell {} Primitive {}: Primitive is unused (all coeffs = 0.0)".format(
                            element_z, idx, pidx))


def _validate_element(el_data, element_z):
    if 'electron_shells' in el_data:
        _validate_electron_shells(el_data['electron_shells'], element_z)

    if 'ecp_potentials' in el_data:
        if not 'ecp_electrons' in el_data:
            raise RuntimeError("ecp_electrons doesn't exist for element {}, but ecp_potentials does".format(element_z))
        _validate_ecp_potentials(el_data['ecp_potentials'], el_data['ecp_electrons'], element_z)


def _validate_extra_component(bs_data):
    '''Extra checks for component basis files'''

    assert len(bs_data['elements']) > 0

    for element_z, el_data in bs_data['elements'].items():
        _validate_element(el_data, element_z)


def _validate_extra_element(bs_data):
    '''Extra checks for basis metadata files'''

    assert len(bs_data['elements']) > 0


def _validate_extra_table(bs_data):
    '''Extra checks for table basis files'''

    assert len(bs_data['elements']) > 0

    # Will throw an exception on invalid dates
    datetime.datetime.strptime(bs_data['revision_date'], "%Y-%m-%d")


def _validate_extra_complete(bs_data):
    '''Extra checks for complete basis set data'''

    assert len(bs_data['elements']) > 0

    # Make sure 'name' exists in the list of 'names'
    if not bs_data['name'] in bs_data['names']:
        raise RuntimeError("Name {} not part of names: {}".format(bs_data['name'], bs_data['names']))

    for element_z, el_data in bs_data['elements'].items():
        _validate_element(el_data, element_z)


def _validate_extra_minimal(bs_data):
    '''Extra checks for minimal basis set data'''

    assert len(bs_data['elements']) > 0

    for element_z, el_data in bs_data['elements'].items():
        _validate_element(el_data, element_z)


_validate_map = {
    'references': _validate_extra_references,
    'metadata': _validate_extra_metadata,
    'component': _validate_extra_component,
    'element': _validate_extra_element,
    'table': _validate_extra_table,
    'complete': _validate_extra_complete,
    'minimal': _validate_extra_minimal
}


def _get_schema(file_type):
    '''Get a schema that can validate BSE JSON files or dictionaries

       The schema_type represents the type of BSE JSON file to be validated,
       and can be 'component', 'element', 'table', 'metadata', or 'references'.

       Returns the schema and the reference resolver
    '''

    schema_file = "{}-schema.json".format(file_type)
    file_path = os.path.join(_default_schema_dir, schema_file)

    schema = fileio.read_schema(file_path)

    # Set up the resolver for links
    base_uri = 'file://{}/'.format(_default_schema_dir)
    resolver = jsonschema.RefResolver(base_uri=base_uri, referrer=schema)

    return schema, resolver


def validate_data(file_type, bs_data):
    """
    Validates json basis set data against a schema

    Parameters
    ----------
    file_type : str
        Type of the data to validate. May be 'component', 'element', 'table', 'complete', 'minimal', or 'references'
    bs_data:
        Data to be validated

    Raises
    ------
    RuntimeError
        If the file_type is not valid (and/or a schema doesn't exist)
    ValidationError
        If the given file does not pass validation
    FileNotFoundError
        If the file given by file_path doesn't exist
    """

    if file_type not in _validate_map:
        raise RuntimeError("{} is not a valid file_type".format(file_type))

    schema, resolver = _get_schema(file_type)
    jsonschema.validate(bs_data, schema, resolver=resolver)
    _validate_map[file_type](bs_data)


def validate_file(file_type, file_path):
    """
    Validates a file against a schema

    Parameters
    ----------
    file_type : str
        Type of file to read. May be 'component', 'element', 'table', 'complete', 'minimal', or 'references'
    file_path:
        Full path to the file to be validated

    Raises
    ------
    RuntimeError
        If the file_type is not valid (and/or a schema doesn't exist)
    ValidationError
        If the given file does not pass validation
    FileNotFoundError
        If the file given by file_path doesn't exist
    """

    file_data = fileio._read_plain_json(file_path, False)
    validate_data(file_type, file_data)


def validate_data_dir(data_dir):
    """
    Validates all files in a data_dir
    """

    all_meta, all_table, all_element, all_component = fileio.get_all_filelist(data_dir)

    for f in all_meta:
        full_path = os.path.join(data_dir, f)
        validate_file('metadata', full_path)
    for f in all_table:
        full_path = os.path.join(data_dir, f)
        validate_file('table', full_path)
    for f in all_element:
        full_path = os.path.join(data_dir, f)
        validate_file('element', full_path)
    for f in all_component:
        full_path = os.path.join(data_dir, f)
        validate_file('component', full_path)



'''
Command line interface for the basis set exchange
'''

import argparse
import argcomplete
from .. import version
from .bse_handlers import bse_cli_handle_subcmd
from .check import cli_check_normalize_args
from .complete import (cli_case_insensitive_validator,
                       cli_family_completer, cli_role_completer, cli_bsname_completer,
                       cli_write_fmt_completer, cli_read_fmt_completer, cli_reffmt_completer)


def run_bse_cli():
    ################################################################################################
    # NOTE: I am deliberately not using the 'choices' argument in add_argument. I could use it
    # for formats, etc, however I wouldn't want to use it for basis set names. Therefore, I handle
    # all of that manually so that error output is consistent and clean
    ################################################################################################

    ########################################
    # Main global options
    ########################################
    parser = argparse.ArgumentParser(description='Description of your program')
    parser.add_argument('-V', action='version', version='basis_set_exchange ' + version())
    parser.add_argument('-d', '--data-dir', metavar='PATH', help='Override which data directory to use')
    parser.add_argument('-o', '--output', metavar='PATH', help='Output to given file rather than stdout')

    subparsers = parser.add_subparsers(metavar='subcommand', dest='subcmd')
    subparsers.required = True # https://bugs.python.org/issue9253#msg186387

    ########################################
    # Listing of data-independent info
    ########################################
    # list-formats subcommand
    subp = subparsers.add_parser('list-formats', help='Output a list of basis set formats that can be used with obtaining a basis set')
    subp.add_argument('-n', '--no-description', action='store_true', help='Print only the format names')

    # list-writer-formats subcommand
    subp = subparsers.add_parser('list-writer-formats', help='Output a list available basis set formats that can be written')
    subp.add_argument('-n', '--no-description', action='store_true', help='Print only the format names')

    # list-reader-formats
    subp = subparsers.add_parser('list-reader-formats', help='Output a list of basis set formats that can be read')
    subp.add_argument('-n', '--no-description', action='store_true', help='Print only the format names')

    # list-ref-formats subcommand
    subp = subparsers.add_parser('list-ref-formats', help='Output a list all available reference formats and descriptions')
    subp.add_argument('-n', '--no-description', action='store_true', help='Print only the reference format names')

    # list-roles subcommand
    subp = subparsers.add_parser('list-roles', help='Output a list all available roles and descriptions')
    subp.add_argument('-n', '--no-description', action='store_true', help='Print only the role names')

    ########################################
    # Listing of general info and metadata
    ########################################
    # get-data-dir
    subparsers.add_parser('get-data-dir', help='Output the default data directory of this package')

    # list-basis-sets subcommand
    subp = subparsers.add_parser('list-basis-sets', help='Output a list all available basis sets and descriptions')
    subp.add_argument('-n', '--no-description', action='store_true', help='Print only the basis set names')
    subp.add_argument('-f', '--family', help='Limit the basis set list to only the specified family').completer = cli_family_completer
    subp.add_argument('-r', '--role', help='Limit the basis set list to only the specified role').completer = cli_role_completer
    subp.add_argument('-s', '--substr', help='Limit the basis set list to only basis sets whose name contains the specified substring')
    subp.add_argument('-e', '--elements', help='Limit the basis set list to only basis sets that contain all the given elements')

    # list-families subcommand
    subparsers.add_parser('list-families', help='Output a list all available basis set families')

    # lookup-by-role
    subp = subparsers.add_parser('lookup-by-role', help='Lookup a companion/auxiliary basis by primary basis and role')
    subp.add_argument('basis', help='Name of the primary basis we want the auxiliary basis for').completer = cli_bsname_completer
    subp.add_argument('role', help='Role of the auxiliary basis to look for').completer = cli_role_completer

    #################################
    # Output of info
    #################################
    # get-basis subcommand
    subp = subparsers.add_parser('get-basis', help='Output a formatted basis set')
    subp.add_argument('basis', help='Name of the basis set to output').completer = cli_bsname_completer
    subp.add_argument('fmt', help='Which format to output the basis set as').completer = cli_write_fmt_completer
    subp.add_argument('--elements', help='Which elements of the basis set to output. Default is all defined in the given basis')
    subp.add_argument('--version', help='Which version of the basis set to output. Default is the latest version')
    subp.add_argument('--noheader', action='store_true', help='Do not output the header at the top')
    subp.add_argument('--unc-gen', action='store_true', help='Remove general contractions')
    subp.add_argument('--unc-spdf', action='store_true', help='Remove combined sp, spd, ... contractions')
    subp.add_argument('--unc-seg', action='store_true', help='Remove general contractions')
    subp.add_argument('--opt-gen', action='store_true', help='Optimize general contractions')
    subp.add_argument('--make-gen', action='store_true', help='Make the basis set as generally-contracted as possible')

    # get-refs subcommand
    subp = subparsers.add_parser('get-refs', help='Output references for a basis set')
    subp.add_argument('basis', help='Name of the basis set to output the references for').completer = cli_bsname_completer
    subp.add_argument('reffmt', help='Which format to output the references as').completer = cli_reffmt_completer
    subp.add_argument('--elements', help='Which elements to output the references for. Default is all defined in the given basis.')
    subp.add_argument('--version', help='Which version of the basis set to get the references for')

    # get-info subcommand
    subp = subparsers.add_parser('get-info', help='Output general info and metadata for a basis set')
    subp.add_argument('basis', help='Name of the basis set to output the info for').completer = cli_bsname_completer

    # get-notes subcommand
    subp = subparsers.add_parser('get-notes', help='Output the notes for a basis set')
    subp.add_argument('basis', help='Name of the basis set to output the notes for').completer = cli_bsname_completer

    # get-family subcommand
    subp = subparsers.add_parser('get-family', help='Output the family of a basis set')
    subp.add_argument('basis', help='Name of the basis set to output the family for').completer = cli_bsname_completer

    # get-versions subcommand
    subp = subparsers.add_parser('get-versions', help='Output a list all available versions of a basis set')
    subp.add_argument('basis', help='Name of the basis set to list the versions of').completer = cli_bsname_completer
    subp.add_argument('-n', '--no-description', action='store_true', help='Print only the version numbers')

    # get-family-notes subcommand
    subp = subparsers.add_parser('get-family-notes', help='Get the notes of a family of basis sets')
    subp.add_argument('family', type=str.lower, help='The basis set family to the get the notes of').completer = cli_family_completer

    #################################
    # Converting basis sets
    #################################
    subp = subparsers.add_parser('convert-basis', help='Convert basis set files from one format to another')
    subp.add_argument('input_file', type=str, help='Basis set file to convert')
    subp.add_argument('output_file', type=str, help='Converted basis set file')
    subp.add_argument('--in-fmt', type=str, default=None, help='Input format (default: autodetected from input filename').completer = cli_read_fmt_completer
    subp.add_argument('--out-fmt', type=str, default=None, help='Output format (default: autodetected from output filename').completer = cli_write_fmt_completer

    #################################
    # Creating bundles
    #################################
    subp = subparsers.add_parser('create-bundle', help='Create a bundle of basis sets')
    subp.add_argument('fmt', help='Which format to output the basis set as').completer = cli_write_fmt_completer
    subp.add_argument('reffmt', help='Which format to output the references as').completer = cli_reffmt_completer
    subp.add_argument('bundle_file', help='Bundle/Archive file to create')
    subp.add_argument('--archive-type', help='Override the type of archive to create (zip or tbz)')

    #############################
    # DONE WITH SUBCOMMANDS
    #############################

    # setup autocomplete
    argcomplete.autocomplete(parser, validator=cli_case_insensitive_validator)

    # Now parse and handle the args
    args = parser.parse_args()

    # Check and make sure basis sets, roles, etc, are valid
    args = cli_check_normalize_args(args)

    # Actually generate the output
    output = bse_cli_handle_subcmd(args)

    if args.output:
        with open(args.output, 'w', encoding='utf-8') as outfile:
            outfile.write(output)
    else:
        print(output)

    return 0



'''
Helpers for printing pieces of basis sets
'''

from . import lut
from .misc import expand_elements, contraction_string


def _find_point(x):
    if isinstance(x, int):
        return 0
    else:
        return x.index('.')


def _determine_leftpad(column, point_place):
    '''Find how many spaces to put before a column of numbers
       so that all the decimal points line up

    This function takes a column of decimal numbers, and returns a
    vector containing the number of spaces to place before each number
    so that (when possible) the decimal points line up.


    Parameters
    ----------
    column : list
        Numbers that will be printed as a column
    point_place : int
        Number of the character column to put the decimal point
    '''

    # Find the number of digits before the decimal
    ndigits_left = [_find_point(x) for x in column]

    # find the padding per entry, filtering negative numbers
    return [max((point_place - 1) - x, 0) for x in ndigits_left]


def write_matrix(mat, point_place, convert_exp=False):

    # Padding for the whole matrix
    pad = [_determine_leftpad(c, point_place[i]) for i, c in enumerate(mat)]

    # Use the transposes (easier to write out by row)
    pad = list(map(list, zip(*pad)))
    mat = list(map(list, zip(*mat)))

    lines = ''
    for r, row in enumerate(mat):
        line = ''
        for c, val in enumerate(row):
            sp = pad[r][c] - len(line)
            # ensure at least one space, except for the beginning of the line
            if c > 0:
                sp = max(sp, 1)
            line += ' ' * sp + str(mat[r][c])
        lines += line + '\n'

    if convert_exp:
        lines = lines.replace('e', 'D')
        lines = lines.replace('E', 'D')

    return lines


def electron_shell_str(shell, shellidx=None):
    '''Return a string representing the data for an electron shell

    If shellidx (index of the shell) is not None, it will also be printed
    '''
    am = shell['angular_momentum']
    amchar = lut.amint_to_char(am)
    amchar = amchar.upper()

    shellidx_str = ''
    if shellidx is not None:
        shellidx_str = 'Index {} '.format(shellidx)

    exponents = shell['exponents']
    coefficients = shell['coefficients']
    ncol = len(coefficients) + 1

    region = shell['region']
    if region == '':
        region = '(none)'

    point_places = [8 * i + 15 * (i - 1) for i in range(1, ncol + 1)]
    s = "Shell: {} Region: {} AM: {}\n".format(shellidx_str, region, amchar)
    s += "Function: {}\n".format(shell['function_type'])
    s += write_matrix([exponents, *coefficients], point_places)
    return s


def ecp_pot_str(pot):
    '''Return a string representing the data for an ECP potential
    '''

    am = pot['angular_momentum']
    amchar = lut.amint_to_char(am)

    rexponents = pot['r_exponents']
    gexponents = pot['gaussian_exponents']
    coefficients = pot['coefficients']

    point_places = [0, 10, 33]
    s = 'Potential: {} potential\n'.format(amchar)
    s += 'Type: {}\n'.format(pot['ecp_type'])
    s += write_matrix([rexponents, gexponents, *coefficients], point_places)
    return s


def element_data_str(z, eldata):
    '''Return a string with all data for an element

    This includes shell and ECP potential data

    Parameters
    ----------
    z : int or str
        Element Z-number
    eldata: dict
        Data for the element to be printed
    '''

    sym = lut.element_sym_from_Z(z, True)

    cs = contraction_string(eldata)
    if cs == '':
        cs = '(no electron shells)'
    s = '\nElement: {} : {}\n'.format(sym, cs)

    if 'electron_shells' in eldata:
        for shellidx, shell in enumerate(eldata['electron_shells']):
            s += electron_shell_str(shell, shellidx) + '\n'

    if 'ecp_potentials' in eldata:
        s += 'ECP: Element: {}   Number of electrons: {}\n'.format(sym, eldata['ecp_electrons'])

        for pot in eldata['ecp_potentials']:
            s += ecp_pot_str(pot) + '\n'

    return s


def component_basis_str(basis, elements=None):
    '''Print a component basis set

    If elements is not None, only the specified elements will be printed
    (see :func:`bse.misc.expand_elements`)
    '''

    s = "Description: " + basis['description'] + '\n'

    eldata = basis['elements']

    # Filter to the given elements
    if elements is None:
        elements = list(eldata.keys())
    else:
        elements = expand_elements(elements, True)

    # Add the str for each element
    for z in elements:
        s += element_data_str(z, eldata[z]) + '\n'

    return s



"""
Compares basis sets with authoritative versions
"""

import os
import pytest

from basis_set_exchange import curate
from .common_testvars import bs_metadata, bs_names, auth_src_map


@pytest.mark.parametrize('basis_name_ver', list(auth_src_map.keys()))
def test_authoritative(basis_name_ver):
    '''
    Compare the stored basis sets with the stored authoritative sources
    '''

    basis_name, ver = os.path.splitext(basis_name_ver)
    ver = ver[1:]  # remove starting '.'

    basis_meta = bs_metadata[basis_name]

    ref_path = auth_src_map[basis_name_ver]

    if basis_name not in bs_names:
        raise RuntimeError("Source basis {} doesn't have a BSE basis".format(basis_name))

    assert curate.compare_basis_against_file(basis_name, ref_path, version=ver, uncontract_general=True)



"""
Tests for the BSE main API
"""

import basis_set_exchange as bse
import pytest
import os

from .common_testvars import bs_write_formats, ref_formats, bs_names_sample

my_formats = [x for x in bs_write_formats if x is not None]

@pytest.mark.parametrize('fmt', bs_write_formats)
def test_get_format_extensions(fmt):
    """For all basis set formats, get the extension
    """
    bse.writers.get_format_extension(fmt)


@pytest.mark.parametrize('fmt', ref_formats)
def test_get_refformat_extensions(fmt):
    """For all basis set formats, get the extension
    """
    bse.refconverters.get_format_extension(fmt)


@pytest.mark.parametrize('bs_name', bs_names_sample)
@pytest.mark.parametrize('bs_fmt', my_formats)
@pytest.mark.parametrize('as_bz2', [True, False])
def test_write_basis_file(bs_name, bs_fmt, as_bz2, tmp_path):
    '''Test writing a basis set to a file'''

    tmp_path = str(tmp_path)  # Needed for python 3.5

    bs_data = bse.get_basis(bs_name)
    out_file = bse.misc.basis_name_to_filename(bs_name)
    out_file += bse.writers.get_format_extension(bs_fmt)

    if as_bz2:
        out_file += '.bz2'

    outfile_path = os.path.join(tmp_path, out_file)
    bse.writers.write_formatted_basis_file(bs_data, outfile_path, bs_fmt)



"""
Tests of memoization
"""

import random
import pytest
import time
import basis_set_exchange as bse

from .common_testvars import *

# Use random for getting sets of elements
random.seed(rand_seed, version=2)


def _test_memo_helper(func, *args, **kwargs):
    start = time.time()
    try1 = func(*args, **kwargs)
    mid = time.time()
    try2 = func(*args, **kwargs)
    end = time.time()

    # Time should be shorter, but is a bit
    # non-determinstic
    #print(end-mid, mid-start)
    #assert (end-mid) < (mid-start)

    # Should be equal, but not aliased
    assert try1 == try2

    # Empty strings are the same
    if try1 == '':
        return

    assert try1 is not try2


def test_get_metadata_memo():
    """Test memoization of get_metadata"""
    _test_memo_helper(bse.get_metadata)


def test_get_reference_data_memo():
    """Test memoization of get_reference_data"""
    _test_memo_helper(bse.get_reference_data)


@pytest.mark.parametrize('basis_name', bs_names_sample)
def test_get_basis_notes_memo(basis_name):
    """Test memoization of get_basis_notes"""
    _test_memo_helper(bse.get_basis_notes, basis_name)


@pytest.mark.parametrize('basis_name', bs_names_sample)
def test_get_basis_memo(basis_name):
    """Test memoization of get_basis_notes"""
    _test_memo_helper(bse.get_basis, basis_name)


@pytest.mark.parametrize('family', all_families)
def test_get_family_notes_memo(family):
    """Test memoization of get_family_notes"""
    _test_memo_helper(bse.get_family_notes, family)



"""
Tests BSE manip functions
"""

import pytest

from basis_set_exchange import api, curate, manip, sort
from .common_testvars import bs_names


@pytest.mark.slow
@pytest.mark.parametrize('basis', bs_names)
def test_manip_roundtrip_slow(basis):
    bse_dict = api.get_basis(basis)
    bse_dict_gen = manip.make_general(bse_dict)
    bse_dict_unc = manip.uncontract_general(bse_dict_gen)
    bse_dict_unc = manip.prune_basis(bse_dict_unc)
    bse_dict_sort = sort.sort_basis(bse_dict_unc)

    bse_dict = manip.uncontract_general(bse_dict)
    bse_dict = manip.uncontract_spdf(bse_dict)
    assert curate.compare_basis(bse_dict, bse_dict_unc, rel_tol=0.0)
    assert curate.compare_basis(bse_dict, bse_dict_sort, rel_tol=0.0)

    bse_dict_gen = manip.prune_basis(bse_dict_gen)
    bse_dict_gen2 = manip.make_general(bse_dict_unc)
    bse_dict_gen2 = manip.prune_basis(bse_dict_gen2)
    assert curate.compare_basis(bse_dict_gen, bse_dict_gen2, rel_tol=0.0)



'''
Test for unused data
'''

import os
import basis_set_exchange as bse
from .common_testvars import data_dir, all_component_paths, all_element_paths, all_table_paths, all_metadata_files, all_families


def test_unused_data():
    '''
    Test for any unused data in the data directory
    '''

    # All elements contained in all component files
    all_component_elements = {}
    for component_path in all_component_paths:
        component_data = bse.fileio.read_json_basis(component_path)
        all_component_elements[component_path] = list(component_data['elements'].keys())

    # All elements contained in all element files
    # And all element data as read from the file
    all_element_elements = {}
    all_element_data = {}
    for element_path in all_element_paths:
        element_data = bse.fileio.read_json_basis(element_path)
        all_element_elements[element_path] = list(element_data['elements'].keys())
        all_element_data[element_path] = element_data['elements']

    # Now go through what is reachable through a table file
    for table_path in all_table_paths:
        table_data = bse.fileio.read_json_basis(table_path)

        # What element files are linked to this table file
        el_files = list(table_data['elements'].items())

        # Loop over the element files, and remove the corresponding entry
        # from all_component_elements
        for el, el_file in el_files:
            # Normalize the paths (since we will be removing them later)
            el_file = os.path.normpath(el_file)
            el_file_path = os.path.join(data_dir, el_file)

            el_file_data = all_element_data[el_file_path]

            for cfile in el_file_data[el]['components']:
                cfile = os.path.normpath(cfile)
                cfile_path = os.path.join(data_dir, cfile)
                if el in all_component_elements[cfile_path]:
                    all_component_elements[cfile_path].remove(el)

            # Now remove the corresponding entry from all_element_elements
            if el in all_element_elements[el_file_path]:
                all_element_elements[el_file_path].remove(el)

    # See which ones were unused
    found_unused = False

    # Merge into one big dictionary
    remaining = all_component_elements
    remaining.update(all_element_elements)

    for k, v in remaining.items():
        if not v:
            continue

        found_unused = True
        for el in v:
            print("Element {:3} in {} not used".format(el, k))

    if found_unused:
        raise RuntimeError("Found unused data")


def test_unused_notes():
    '''
    Test for orphan basis and family notes files
    '''

    all_basis_notes = []
    all_family_notes = []
    for root, dirs, files in os.walk(data_dir):
        for basename in files:
            fpath = os.path.join(root, basename)
            fpath = os.path.relpath(fpath, data_dir)

            if basename.endswith('.notes'):
                all_basis_notes.append(fpath)
            elif basename.startswith('NOTES.'):
                all_family_notes.append(fpath)

    found_unused = False
    for bs_notes in all_basis_notes:
        base = os.path.splitext(bs_notes)[0]
        metafile = base + '.metadata.json'
        if metafile not in all_metadata_files:
            print("File {} does not have a corresponding metadata file".format(bs_notes))
            found_unused = True

    for fam_notes in all_family_notes:
        fam = os.path.splitext(fam_notes)[1][1:] # Removes period
        if fam not in all_families:
            print("File {} does not have a corresponding family".format(fam_notes))
            found_unused = True

    if found_unused:
        raise RuntimeError("Found unused notes files")



'''
Conversion of basis sets to GAMESS-UK format
'''

from .. import lut, manip, sort, printing


def write_gamess_uk(basis):
    '''Converts a basis set to GAMESS-UK
    '''

    s = ''

    # Uncontract all but SP
    basis = manip.uncontract_general(basis, True)
    basis = manip.uncontract_spdf(basis, 1, False)
    basis = sort.sort_basis(basis, False)

    # Elements for which we have electron basis
    electron_elements = [k for k, v in basis['elements'].items() if 'electron_shells' in v]

    # Elements for which we have ECP
    ecp_elements = [k for k, v in basis['elements'].items() if 'ecp_potentials' in v]

    # Electron Basis
    if electron_elements:
        # electronic part starts with $DATA

        for z in electron_elements:
            data = basis['elements'][z]

            el_name = lut.element_name_from_Z(z).upper()
            el_sym = lut.element_sym_from_Z(z, normalize=True)
            s += '\n'
            s += '# ' + el_name + "\n"

            for shell in data['electron_shells']:
                exponents = shell['exponents']
                coefficients = shell['coefficients']
                ncol = len(coefficients) + 2  #include index column

                am = shell['angular_momentum']
                amchar = lut.amint_to_char(am, hij=True, use_L=True).upper()
                s += '{}   {}\n'.format(amchar, el_sym)

                # 1-based indexing
                point_places = [8 * i + 15 * (i - 1) for i in range(1, ncol)]

                # Note: order for sp shells is (coeff of s) (exponents) (coeff of p)
                s += printing.write_matrix([coefficients[0], exponents, *coefficients[1:]], point_places)

    # Write out ECP
    if ecp_elements:
        s += "\n\nEffective Core Potentials\n"
        s += "---------------------------\n"

        for z in ecp_elements:
            data = basis['elements'][z]
            sym = lut.element_sym_from_Z(z).upper()
            max_ecp_am = max([x['angular_momentum'][0] for x in data['ecp_potentials']])

            # Sort lowest->highest, then put the highest at the beginning
            ecp_list = sorted(data['ecp_potentials'], key=lambda x: x['angular_momentum'])
            ecp_list.insert(0, ecp_list.pop())

            s += 'CARDS {}\n'.format(sym)
            s += '    {}     {}\n'.format(max_ecp_am, data['ecp_electrons'])

            for pot in ecp_list:
                rexponents = pot['r_exponents']
                gexponents = pot['gaussian_exponents']
                coefficients = pot['coefficients']

                point_places = [1, 9, 32]
                s += printing.write_matrix([rexponents, *coefficients, gexponents], point_places)
            s += '\n'
    return s



"""
Validation of bibtex output
"""

import pytest
import json
from basis_set_exchange import api, lut
from .common_testvars import bs_names_sample

_has_qcschema = True
try:
    import qcschema
except ImportError:
    _has_qcschema = False

# yapf: disable
@pytest.mark.skipif(not _has_qcschema, reason="QCSchema not available to test qcschema output")
@pytest.mark.parametrize('basis_name', bs_names_sample)
# yapf: enable
def test_valid_qcschema(basis_name):
    basis_dict = api.get_basis(basis_name)
    qcs_str = api.get_basis(basis_name, fmt='qcschema')
    qcs_json = json.loads(qcs_str)

    el_list = [lut.element_sym_from_Z(x, True) for x in basis_dict['elements'].keys()]
    coords = []
    for idx, el in enumerate(el_list):
        coords.extend((0.0, 0.0, float(idx)))

    qcs_json['atom_map'] = list(qcs_json['center_data'].keys())
    assert len(qcs_json['atom_map']) == len(el_list)

    dummy_inp = {
        "schema_name": "qc_schema_input",
        "schema_version": 1,
        "keywords": {},
        "molecule": {
          "schema_name": "qcschema_molecule",
          "schema_version": 2,
          "geometry": coords,
          "symbols": el_list
        },
        'driver': 'energy',
        'model': {
            'method': 'B3LYP',
            'basis': qcs_json
        }
    }

    qcschema.validate(dummy_inp, 'input')



"""
Test for validating the formatting of the json files
"""

import glob
import os
import pytest

from basis_set_exchange import validator
from .common_testvars import data_dir


def test_valid_data_dir():
    validator.validate_data_dir(data_dir)



'''
Conversion of basis sets to BDF format
'''

from .. import lut, manip, printing, sort, misc


def write_bdf(basis):
    '''Converts a basis set to BDF format
    '''

    basis = manip.make_general(basis, False, True)
    basis = sort.sort_basis(basis, False)

    s = ''

    # Elements for which we have electron basis
    electron_elements = [k for k, v in basis['elements'].items() if 'electron_shells' in v]

    # Elements for which we have ECP
    ecp_elements = [k for k, v in basis['elements'].items() if 'ecp_potentials' in v]

    if electron_elements:
        # Electron Basis
        for z in electron_elements:
            s += '****\n'
            data = basis['elements'][z]
            #Get element symbol
            s += lut.element_sym_from_Z(z, True)

            max_am = misc.max_am(data['electron_shells'])
            s += '{:>7}   {}\n'.format(z, max_am)

            for shell in data['electron_shells']:
                exponents = shell['exponents']
                coefficients = shell['coefficients']
                nprim = len(exponents)
                ngen = len(coefficients)

                amchar = lut.amint_to_char(shell['angular_momentum']).upper()
                s += '{}    '.format(amchar)
                s += '{:>3}    {}\n'.format(nprim, ngen)

                s += printing.write_matrix([exponents], [17])

                point_places = [8 * i + 15 * (i - 1) for i in range(1, ngen + 1)]
                s += printing.write_matrix(coefficients, point_places)

    # Write out ECP
    if ecp_elements:
        s += '\n\nECP\n'

        for z in ecp_elements:
            data = basis['elements'][z]
            sym = lut.element_sym_from_Z(z, True)
            max_ecp_am = max([x['angular_momentum'][0] for x in data['ecp_potentials']])

            # Sort lowest->highest, then put the highest at the beginning
            ecp_list = sorted(data['ecp_potentials'], key=lambda x: x['angular_momentum'])
            ecp_list.insert(0, ecp_list.pop())

            s += '{} nelec {}\n'.format(sym, data['ecp_electrons'])

            for pot in ecp_list:
                rexponents = pot['r_exponents']
                gexponents = pot['gaussian_exponents']
                coefficients = pot['coefficients']

                am = pot['angular_momentum']
                amchar = lut.amint_to_char(am).upper()

                if am[0] == max_ecp_am:
                    s += '{} ul\n'.format(sym)
                else:
                    s += '{} {}\n'.format(sym, amchar)

                point_places = [0, 10, 33]
                s += printing.write_matrix([rexponents, gexponents, *coefficients], point_places)

        s += 'END\n'

    return s



'''
Validators for command line options
'''

import os
import copy
from .. import api, misc, curate


def _cli_check_data_dir(data_dir):
    '''Checks that the data dir exists and contains METADATA.json'''

    if data_dir is None:
        return None

    data_dir = os.path.expanduser(data_dir)
    data_dir = os.path.expandvars(data_dir)
    if not os.path.isdir(data_dir):
        raise RuntimeError("Data directory '{}' does not exist or is not a directory".format(data_dir))
    if not os.path.isfile(os.path.join(data_dir, 'METADATA.json')):
        raise RuntimeError("Data directory '{}' does not contain a METADATA.json file".format(data_dir))

    return data_dir


def _cli_check_format(fmt):
    '''Checks that a basis set format exists and if not, raises a helpful exception'''

    if fmt is None:
        return None

    fmt = fmt.lower()
    if fmt not in api.get_formats():
        errstr = "Format '" + fmt + "' does not exist.\n"
        errstr += "For a complete list of formats, use the 'bse list-formats' command"
        raise RuntimeError(errstr)

    return fmt


def _cli_check_ref_format(fmt):
    '''Checks that a reference format exists and if not, raises a helpful exception'''

    if fmt is None:
        return None

    fmt = fmt.lower()
    if fmt not in api.get_reference_formats():
        errstr = "Reference format '" + fmt + "' does not exist.\n"
        errstr += "For a complete list of formats, use the 'bse list-ref-formats' command"
        raise RuntimeError(errstr)

    return fmt


def _cli_check_role(role):
    '''Checks that a basis set role exists and if not, raises a helpful exception'''

    if role is None:
        return None

    role = role.lower()
    if role not in api.get_roles():
        errstr = "Role format '" + role + "' does not exist.\n"
        errstr += "For a complete list of roles, use the 'bse list-roles' command"
        raise RuntimeError(errstr)

    return role


def _cli_check_basis(name, data_dir):
    '''Checks that a basis set exists and if not, raises a helpful exception'''

    if name is None:
        return None

    name = misc.transform_basis_name(name)
    metadata = api.get_metadata(data_dir)
    if name not in metadata:
        errstr = "Basis set '" + name + "' does not exist.\n"
        errstr += "For a complete list of basis sets, use the 'bse list-basis-sets' command"
        raise RuntimeError(errstr)

    return name


def _cli_check_family(family, data_dir):
    '''Checks that a basis set family exists and if not, raises a helpful exception'''

    if family is None:
        return None

    family = family.lower()
    if family not in api.get_families(data_dir):
        errstr = "Basis set family '" + family + "' does not exist.\n"
        errstr += "For a complete list of families, use the 'bse list-families' command"
        raise RuntimeError(errstr)

    return family


def _cli_check_readfmt(readfmt):
    '''Checks that a file type exists and if not, raises a helpful exception'''

    if readfmt is None:
        return None

    readfmt = readfmt.lower()
    if readfmt not in curate.get_reader_formats():
        errstr = "Reader for file type '" + readfmt + "' does not exist.\n"
        errstr += "For a complete list of file types, use the 'bsecurate get-reader-formats' command"
        raise RuntimeError(errstr)

    return readfmt


def cli_check_normalize_args(args):
    '''Check and normalize arguments
       This function checks that basis set names, families, roles, etc, are
       valid (and raise an exception if they aren't)

       The original data passed to this function is not modified. A modified
       copy is returned.
    '''

    args_keys = vars(args).keys()  # What args we have
    args_copy = copy.copy(args)
    if 'data_dir' in args_keys:
        args_copy.data_dir = _cli_check_data_dir(args.data_dir)
    if 'basis' in args:
        args_copy.basis = _cli_check_basis(args.basis, args.data_dir)
    if 'basis1' in args_keys:
        args_copy.basis1 = _cli_check_basis(args.basis1, args.data_dir)
    if 'basis2' in args_keys:
        args_copy.basis2 = _cli_check_basis(args.basis2, args.data_dir)
    if 'fmt' in args_keys:
        args_copy.fmt = _cli_check_format(args.fmt)
    if 'reffmt' in args_keys:
        args_copy.reffmt = _cli_check_ref_format(args.reffmt)
    if 'role' in args_keys:
        args_copy.role = _cli_check_role(args.role)
    if 'family' in args_keys:
        args_copy.family = _cli_check_family(args.family, args.data_dir)
    if 'readfmt1' in args_keys:
        args_copy.readfmt1 = _cli_check_readfmt(args.readfmt1)
    if 'readfmt2' in args_keys:
        args_copy.readfmt2 = _cli_check_readfmt(args.readfmt2)

    return args_copy



'''
Class/decorator for memoizing BSE functionality
'''

import functools
import pickle
import inspect

# If set to True, memoization of some internal functions
# will be used. Generally safe to leave enabled - it
# won't use that much memory
memoize_enabled = True


def _make_key(args_spec, *args, **kwargs):
    left_args = args_spec.args[len(args):]
    num_defaults = len(args_spec.defaults or ())
    defaults_names = args_spec.args[-num_defaults:]

    if not set(left_args).symmetric_difference(kwargs).issubset(defaults_names):
        # We got an error in the function call. Let's simply trigger it
        func(*args, **kwargs)

    start = 0
    key = []
    for arg, arg_name in zip(args, args_spec.args):
        key.append(arg)
        if arg_name in defaults_names:
            start += 1

    for left_arg in left_args:
        try:
            key.append(kwargs[left_arg])
        except KeyError:
            key.append(args_spec.defaults[start])

        # Increase index if we used a default, or if the argument was provided
        if left_arg in defaults_names:
            start += 1

    return pickle.dumps(key)


class BSEMemoize:
    def __init__(self, f):
        self.__f = f
        self.args_spec = inspect.getfullargspec(f)
        self.__memo = {}
        functools.update_wrapper(self, f)

    def __call__(self, *args, **kwargs):
        if not memoize_enabled:
            return self.__f(*args, **kwargs)

        arg_key = _make_key(self.args_spec, *args, **kwargs)

        if arg_key in self.__memo:
            return pickle.loads(self.__memo[arg_key])

        ret = self.__f(*args, **kwargs)
        self.__memo[arg_key] = pickle.dumps(ret)
        return ret



'''
Conversion of basis sets to PQS
'''

from .. import lut, manip, sort, printing
from .gamess_us import write_gamess_us_ecp_basis


def write_pqs_electron_basis(basis, electron_elements):
    s = ''

    for z in electron_elements:
        data = basis['elements'][z]

        el_sym = lut.element_sym_from_Z(z, normalize=True)
        s += 'FOR        ' + el_sym + "\n"

        for shell in data['electron_shells']:
            exponents = shell['exponents']
            coefficients = shell['coefficients']

            am = shell['angular_momentum']
            amchar = lut.amint_to_char(am, hij=True, use_L=True).upper()

            ncol = len(coefficients) + 1
            point_places = [4 + 8 * i + 15 * (i - 1) for i in range(1, ncol+1)]
            mat = printing.write_matrix([exponents, *coefficients], point_places)

            # Prepend the AM
            mat = amchar + mat[1:]
            s += mat

    return s


def write_pqs(basis):
    '''Converts the basis set to PQS
    '''

    s = ''

    basis = manip.make_general(basis, True)
    basis = sort.sort_basis(basis, False)

    # Elements for which we have electron basis
    electron_elements = [k for k, v in basis['elements'].items() if 'electron_shells' in v]

    # Elements for which we have ECP
    ecp_elements = [k for k, v in basis['elements'].items() if 'ecp_potentials' in v]

    # Electron Basis
    if electron_elements:
        s += write_pqs_electron_basis(basis, electron_elements)

    # Write out ECP
    if ecp_elements:
        s += '\n\n'
        s += 'Effective core Potentials\n'
        s += '-------------------------\n'
        s += write_gamess_us_ecp_basis(basis, ecp_elements, ecp_block=False)

    return s



import re
from .. import lut, misc
from . import helpers
from .turbomole import _parse_ecp_potential_lines

element_block_re = re.compile(r'^([a-zA-Z]{1,3}):(.*)$')
ecp_block_re = re.compile(r'ncore\s*=\s*(\d+)\s+lmax\s*=\s*(\d+)\s*$', flags=re.IGNORECASE)


def _parse_electron_lines(basis_lines, bs_data):
    # Line 0: element, basis name
    # Line 1: comment
    # Line 2: blank
    # Line 3: nshell (actually line 2 now)

    element_sym, _ = helpers.parse_line_regex(element_block_re, basis_lines[0])
    element_Z = lut.element_Z_from_sym(element_sym, as_str=True)

    element_data = helpers.create_element_data(bs_data, element_Z, 'electron_shells')

    basis_lines = basis_lines[2:]

    # Line 2 (now line 0) should be blank
    basis_lines = helpers.remove_expected_line(basis_lines)

    nshell = helpers.parse_line_regex(r'^(\d+)$', basis_lines[0], "Nshell integer")

    # Read in
    # 1. AM for each shell
    # 2. Nprim for each shell
    # These function calls strip off lines that have already been read
    shell_ams, basis_lines = helpers.read_n_integers(basis_lines[1:], nshell, True)
    shell_ngens, basis_lines = helpers.read_n_integers(basis_lines, nshell, True)
    shell_nprims, basis_lines = helpers.read_n_integers(basis_lines, nshell, True)

    # Loop over all shells
    for shell_idx in range(nshell):

        shell_am = [shell_ams[shell_idx]]
        nprim = shell_nprims[shell_idx]
        ngen = shell_ngens[shell_idx]

        func_type = helpers.function_type_from_am(shell_am, 'gto', 'spherical')

        # Read in exponents and coefficients
        # We know the dimensions of the coefficient matrix. That matrix
        # also has rows which may span multiple columns

        # The next line should be blank (before exponents)
        basis_lines = helpers.remove_expected_line(basis_lines)
        exponents, basis_lines = helpers.read_n_floats(basis_lines, nprim)

        # Next line should be blank (between exponents and coefficients)
        basis_lines = helpers.remove_expected_line(basis_lines)

        coefficients, basis_lines = helpers.parse_fixed_matrix(basis_lines, nprim, ngen)
        coefficients = misc.transpose_matrix(coefficients)

        shell = {
            'function_type': func_type,
            'region': '',
            'angular_momentum': shell_am,
            'exponents': exponents,
            'coefficients': coefficients
        }

        element_data['electron_shells'].append(shell)

    # Remaining lines should either be blank or '*'
    basis_lines = helpers.prune_lines(basis_lines, '*')
    if basis_lines:
        raise RuntimeError("Found extra lines after element block: " + str(basis_lines))


def _parse_ecp_lines(basis_lines, bs_data):
    # Line 0: element, basis name
    # Line 1: comment, but starts with #, so it was removed
    # Line 2: *
    # Line 3: ncore lmax
    #element_sym, _ = helpers.parse_line_regex(element_block_re, basis_lines[0])
    #element_Z = str(lut.element_Z_from_sym(element_sym))

    #element_data = helpers.create_element_data(bs_data, element_Z, 'ecp_potentials')

    # We can use the turbomole parser for ECP.
    # However, needs to be fixed
    # We need to remove the * line
    basis_lines = helpers.remove_expected_line(basis_lines, '*', 1)

    # Also, remove all the other '*' lines (genbas seems to have some extra)
    basis_lines = helpers.prune_lines(basis_lines, '*')

    # Replace the colon on the first line with a space
    basis_lines[0] = basis_lines[0].replace(':', ' ', 1)

    # Now we can use the turbomole version
    _parse_ecp_potential_lines(basis_lines, bs_data)


def read_genbas(basis_lines):
    '''Reads genbas-formatted file data and converts it to a dictionary with the
       usual BSE fields

       This also handles cfour and acesII, which differ only
       in how floating-point numbers are written

       Note that the genbas format does not store all the fields we
       have, so some fields are left blank
    '''

    # We leave in blank lines - they are significant
    # Leave strip_end_blanks to True though
    basis_lines = helpers.prune_lines(basis_lines, '!#', prune_blank=False)

    bs_data = {}

    # split into element blocks
    # each block may be electron shells or ECP
    element_blocks = helpers.partition_lines(basis_lines, element_block_re.match, min_after=1, min_size=4)

    for element_lines in element_blocks:
        if any(ecp_block_re.match(l) for l in element_lines):
            _parse_ecp_lines(element_lines, bs_data)
        else:
            _parse_electron_lines(element_lines, bs_data)

    return bs_data



#!/usr/bin/env python3

import sys
import basis_set_exchange as bse

flist = [
'aug-cc-pV5Z',
'aug-cc-pV5Z-PP',
'aug-cc-pVDZ',
'aug-cc-pVDZ-PP',
'aug-cc-pVQZ',
'aug-cc-pVQZ-PP',
'aug-cc-pVTZ',
'aug-cc-pVTZ-PP',
'aug-cc-pwCV5Z-PP',
'aug-cc-pwCVDZ-PP',
'aug-cc-pwCVQZ-PP',
'aug-cc-pwCVTZ-PP',
'cc-pCVDZ-F12',
'cc-pCVQZ-F12',
'cc-pCVTZ-F12',
'cc-pVDZ-F12',
'cc-pVQZ-F12',
'cc-pVTZ-F12' ]

for bs in flist:
    f = bs + '.metadata.json'
    a = bse.fileio.read_json_basis(f)
    a['auxiliaries']['optri'] = bs.lower() + '-optri'
    bse.fileio.write_json_basis(f, a)



'''
Functionality for handling basis set and family notes
'''

from . import references


def process_notes(notes, ref_data):
    '''Add reference information to the bottom of a notes file

    `:ref:` tags are removed and the actual reference data is appended
    '''

    ref_keys = ref_data.keys()

    found_refs = set()
    for k in ref_keys:
        if k in notes:
            found_refs.add(k)

    # The block to append
    reference_sec = '\n\n'
    reference_sec += '-------------------------------------------------\n'
    reference_sec += ' REFERENCES MENTIONED ABOVE\n'
    reference_sec += ' (not necessarily references for the basis sets)\n'
    reference_sec += '-------------------------------------------------\n'

    # Add reference data
    if not found_refs:
        return notes

    for r in sorted(found_refs):
        rtxt = references.reference_text(r, ref_data[r])
        reference_sec += rtxt + '\n\n'

    return notes + reference_sec



"""
Tests for sanity of auxiliary basis sets
"""

import pytest

from .common_testvars import bs_names, all_roles, bs_metadata


@pytest.mark.parametrize('basis_name', bs_names)
def test_roles(basis_name):
    """Make sure all basis sets have a role that is in api.get_roles()
    """

    bs_role = bs_metadata[basis_name]['role']
    assert bs_role in all_roles



#!/usr/bin/env python3

import os
import sys
import basis_set_exchange as bse

meta, table, _, _ = bse.fileio.get_all_filelist(bse.get_data_dir())


for t in table:
    subdirs = set()
    td = bse.fileio.read_json_basis(t)
    for v in td['elements'].values():
        subdirs.add(os.path.split(v)[0])

    if len(subdirs) > 1:
        print("File: {} has more than one subdir. You are on your own".format(t))
        continue
    assert len(subdirs) > 0
    subdir = subdirs.pop()

    # Find metadata file
    tbase = t.split('.')[0]

    matching_meta = [x for x in meta if x.startswith(tbase + '.')]
    if len(matching_meta) != 1:
        raise RuntimeError("matching meta != 1")
    matching_meta = matching_meta[0]
    print(tbase)
    print('   into ' + subdir)
    print(t)
    print('   ' + matching_meta)



"""
Tests validation of basis set data
"""

import bz2
import os
import json
import pytest

from basis_set_exchange import validator
from .common_testvars import validator_test_data_dir

test_files = os.listdir(validator_test_data_dir)
test_files = [x for x in test_files if x.endswith('.bz2')]

@pytest.mark.parametrize('file_rel_path', test_files)
def test_validator(file_rel_path):
    file_path = os.path.join(validator_test_data_dir, file_rel_path)
    is_bad = ".bad." in os.path.basename(file_path)
    file_type = os.path.basename(file_path).split('.')[1]

    with bz2.open(file_path, 'rt', encoding='utf-8') as tf:
        file_data = tf.read()
    
    if is_bad:
        # Read the first line of the file. This will contain the string
        # that should match the exception
        file_data_lines = file_data.splitlines()
        match = file_data_lines[0][8:].strip()
        json_data = json.loads('\n'.join(file_data_lines[1:]))

        with pytest.raises(Exception, match=match):
            validator.validate_data(file_type, json_data)

    else:
        json_data = json.loads(file_data)
        validator.validate_data(file_type, json_data)



'''
Handlers for command line subcommands
'''

from .. import curate, printing, fileio
from .common import format_columns


def _bsecurate_cli_elements_in_files(args):
    '''Handles the elements-in-files subcommand'''
    data = curate.elements_in_files(args.files)
    return '\n'.join(format_columns(data.items()))


def _bsecurate_cli_component_file_refs(args):
    '''Handles the component-file-refs subcommand'''
    data = curate.component_file_refs(args.files)

    s = ''

    for cfile, cdata in data.items():
        s += cfile + '\n'
        rows = []
        for el, refs in cdata:
            rows.append(('    ' + el, ' '.join(refs)))
        s += '\n'.join(format_columns(rows)) + '\n\n'

    return s


def _bsecurate_cli_print_component_file(args):
    '''Handles the print-component-file subcommand'''

    data = fileio.read_json_basis(args.file)
    return printing.component_basis_str(data, elements=args.elements)


def _bsecurate_cli_compare_basis_sets(args):
    '''Handles compare-basis-sets subcommand'''
    ret = curate.compare_basis_sets(args.basis1, args.basis2, args.version1, args.version2, args.uncontract_general,
          args.data_dir, args.data_dir)
    if ret:
        return "No difference found"
    else:
        return "DIFFERENCES FOUND. SEE ABOVE"


def _bsecurate_cli_compare_basis_files(args):
    '''Handles compare-basis-files subcommand'''
    ret = curate.compare_basis_files(args.file1, args.file2, args.readfmt1, args.readfmt2, args.uncontract_general)

    if ret:
        return "No difference found"
    else:
        return "DIFFERENCES FOUND. SEE ABOVE"


def _bsecurate_cli_make_diff(args):
    '''Handles the view-graph subcommand'''

    curate.diff_json_files(args.left, args.right)
    return ''


def _bsecurate_cli_view_graph(args):
    '''Handles the view-graph subcommand'''

    curate.view_graph(args.basis, args.version, args.data_dir)
    return ''


def _bsecurate_cli_make_graph_file(args):
    '''Handles the make-graph-file subcommand'''

    curate.make_graph_file(args.basis, args.outfile, args.render, args.version, args.data_dir)
    return ''


def bsecurate_cli_handle_subcmd(args):
    handler_map = {
        'elements-in-files': _bsecurate_cli_elements_in_files,
        'component-file-refs': _bsecurate_cli_component_file_refs,
        'print-component-file': _bsecurate_cli_print_component_file,
        'compare-basis-sets': _bsecurate_cli_compare_basis_sets,
        'compare-basis-files': _bsecurate_cli_compare_basis_files,
        'make-diff': _bsecurate_cli_make_diff,
        'view-graph': _bsecurate_cli_view_graph,
        'make-graph-file': _bsecurate_cli_make_graph_file
    }

    return handler_map[args.subcmd](args)



'''
Test for duplicate data in a basis set
'''

import os
import pytest
import basis_set_exchange as bse
from basis_set_exchange import readers
from .common_testvars import bs_names_vers, curate_test_data_dir


def _test_duplicates(bs_dict, expected):
    '''
    Test for any duplicate data in a basis set
    '''

    found_dupe = False
    for el, eldata in bs_dict['elements'].items():
        if 'electron_shells' in eldata:
            # Quick and dirty
            shells = eldata['electron_shells']

            for sh in shells:
                if shells.count(sh) != 1:
                    print("Shell Dupe: element {}".format(el))
                    found_dupe = True
                    break

        if 'ecp_potentials' in eldata:
            pots = eldata['ecp_potentials']

            for pot in pots:
                if pots.count(pot) != 1:
                    print("ECP Dupe: element {}".format(el))
                    found_dupe = True
                    break

    assert found_dupe == expected


@pytest.mark.parametrize('bs_name,bs_ver', bs_names_vers)
def test_duplicate_data(bs_name, bs_ver):
    '''
    Test for any duplicate data in a basis set
    '''

    bs_dict = bse.get_basis(bs_name, version=bs_ver)
    _test_duplicates(bs_dict, False)


@pytest.mark.parametrize('filename', ['6-31g-bse-DUPE.nw.bz2', 'def2-ecp-DUPE.nw.bz2'])
def test_duplicate_fail(filename):
    filepath = os.path.join(curate_test_data_dir, filename)
    filedata = readers.read_formatted_basis_file(filepath, 'nwchem')
    _test_duplicates(filedata, True)



'''
Computing the difference between basis sets and files
'''

import copy
from .compare import compare_electron_shells
from .. import fileio


def subtract_electron_shells(s1, s2, rel_tol=0.0):
    """
    Returns the difference between two lists of electron shells (s1 - s2)

    This will remove any shells from s1 that are also in s2, within a tolerance
    """

    diff_shells = []
    for sh1 in s1:
        for sh2 in s2:
            if compare_electron_shells(sh1, sh2, rel_tol=rel_tol):
                break
        else:
            diff_shells.append(copy.deepcopy(sh1))

    return diff_shells


def diff_basis_dict(left_list, right_list):
    '''
    Compute the difference between two sets of basis set dictionaries

    The result is a list of dictionaries that correspond to each dictionary in
    `left_list`. Each resulting dictionary will contain only the elements/shells
    that exist in that entry and not in any of the dictionaries in `right_list`.

    This only works on the shell level, and will only subtract entire shells
    that are identical. ECP potentials are not affected.

    The return value contains deep copies of the input data

    Parameters
    ----------
    left_list : list of dict
        Dictionaries to use as the base
    right_list : list of dict
        Dictionaries of basis data to subtract from each dictionary of `left_list`

    Returns
    ----------
    list
        Each object in `left_list` containing data that does not appear in `right_list`
    '''

    ret = []
    for bs1 in left_list:
        res = copy.deepcopy(bs1)
        for bs2 in right_list:
            for el in res['elements'].keys():
                if el not in bs2['elements']:
                    continue  # Element only exist in left

                eldata1 = res['elements'][el]
                eldata2 = bs2['elements'][el]

                s1 = eldata1['electron_shells']
                s2 = eldata2['electron_shells']
                eldata1['electron_shells'] = subtract_electron_shells(s1, s2)

        # Remove any empty elements
        res['elements'] = {k: v for k, v in res['elements'].items() if v['electron_shells']}
        ret.append(res)

    return ret


def diff_json_files(left_files, right_files):
    '''
    Compute the difference between two sets of basis set JSON files

    The output is a set of files that correspond to each file in
    `left_files`. Each resulting dictionary will contain only the elements/shells
    that exist in that entry and not in any of the files in `right_files`.

    This only works on the shell level, and will only subtract entire shells
    that are identical. ECP potentials are not affected.

    `left_files` and `right_files` are lists of file paths. The output
    is written to files with the same names as those in `left_files`,
    but with `.diff` added to the end. If those files exist, they are overwritten.

    Parameters
    ----------
    left_files : list of str
        Paths to JSON files to use as the base
    right_files : list of str
        Paths to JSON files to subtract from each file of `left_files`

    Returns
    ----------
    None
    '''

    left_data = [fileio.read_json_basis(x) for x in left_files]
    right_data = [fileio.read_json_basis(x) for x in right_files]
    d = diff_basis_dict(left_data, right_data)

    for idx, diff_bs in enumerate(d):
        fpath = left_files[idx]
        fileio.write_json_basis(fpath + '.diff', diff_bs)



"""
Tests for the BSE IO functions
"""

# Most functionality is covered under other tests.
# This tests the remainder

import os
import pytest

from basis_set_exchange import fileio
from .common_testvars import data_dir


# yapf: disable
@pytest.mark.parametrize('file_path', ['cc-pVDZ.0.table.json',
                                       'CRENBL.0.table.json',
                                       'dunning/cc-pVDZ.1.element.json',
                                       'crenb/CRENBL.0.element.json',
                                       'dunning/cc-pVDZ.1.json',
                                       'crenb/CRENBL.0.json',
                                       'crenb/CRENBL-ECP.0.json'])
# yapf: enable
def test_read_write_basis(file_path):
    # needed to be tested to make sure something isn't left
    # out of the sort lists, etc
    full_path = os.path.join(data_dir, file_path)
    full_path_new = full_path + '.new'
    data = fileio.read_json_basis(full_path)
    fileio.write_json_basis(full_path_new, data)
    os.remove(full_path_new)


@pytest.mark.parametrize('file_path', ['REFERENCES.json'])
def test_read_write_references(file_path):
    # needed to be tested to make sure something isn't left
    # out of the sort lists, etc
    full_path = os.path.join(data_dir, file_path)
    full_path_new = full_path + '.new'
    data = fileio.read_references(full_path)
    fileio.write_references(full_path_new, data)
    os.remove(full_path_new)



'''
Read a basis set file in a given format
'''

import os
import bz2
from ..skel import create_skel
from ..validator import validate_data
from ..compose import _whole_basis_types
from .turbomole import read_turbomole
from .g94 import read_g94
from .nwchem import read_nwchem
from .gbasis import read_gbasis
from .dalton import read_dalton
from .molcas import read_molcas
from .genbas import read_genbas

_reader_map = {
    'turbomole': {
        'display': 'Turbomole',
        'extension': '.tm',
        'reader': read_turbomole
    },
    'gaussian94': {
        'display': 'Gaussian94',
        'extension': '.gbs',
        'reader': read_g94
    },
    'nwchem': {
        'display': 'NWChem',
        'extension': '.nw',
        'reader': read_nwchem
    },
    'dalton': {
        'display': 'Dalton',
        'extension': '.mol',
        'reader': read_dalton
    },
    'molcas': {
        'display': 'MolCAS',
        'extension': '.molcas',
        'reader': read_molcas
    },
    'cfour': {
        'display': 'CFOUR',
        'extension': '.c4bas',
        'reader': read_genbas
    },
    'genbas': {
        'display': 'Genbas',
        'extension': '.genbas',
        'reader': read_genbas
    },
    'gbasis': {
        'display': 'GBasis',
        'extension': '.gbasis',
        'reader': read_gbasis
    }
}


def _fix_uncontracted(basis):
    '''
    Forces the contraction coefficient of uncontracted shells to 1.0
    '''

    for el in basis['elements'].values():
        if 'electron_shells' not in el:
            continue

        for sh in el['electron_shells']:
            if len(sh['coefficients']) == 1 and len(sh['coefficients'][0]) == 1:
                sh['coefficients'][0][0] = '1.0000000'

            # Some uncontracted shells don't have a coefficient
            if len(sh['coefficients']) == 0:
                sh['coefficients'].append(['1.0000000'])

    return basis


def read_formatted_basis_str(basis_str, basis_fmt, validate=False, as_component=False):
    basis_lines = [x.strip() for x in basis_str.splitlines()]

    element_data = _reader_map[basis_fmt]['reader'](basis_lines)

    # the readers give the 'elements' member of a basis set json
    # We need to do a little fixing up
    if as_component:
        data = create_skel('component')
        data['elements'] = element_data
        for el in data['elements'].values():
            el['references'] = []
        bs_type = 'component'
    else:
        data = create_skel('minimal')
        data['elements'] = element_data
        data['name'] = 'unknown_basis'
        data['description'] = 'no_description'
        bs_type = 'minimal'

        # Create the function types
        data['function_types'] = _whole_basis_types(data)

    # It's debateable if I want to do this
    #return _fix_uncontracted(data)

    # Validate if desired
    if validate:
        validate_data(bs_type, data)

    return data


def read_formatted_basis_file(file_path, basis_fmt=None, encoding='utf-8-sig', validate=False, as_component=False):
    # Note that the default is utf-8-sig, which handles the optional byte order mark

    if not os.path.isfile(file_path):
        raise RuntimeError('Basis file path \'{}\' does not exist'.format(file_path))

    if basis_fmt is None:
        for k, v in _reader_map.items():
            ext = v['extension']
            ext_bz2 = ext + '.bz2'
            if file_path.endswith(ext) or file_path.endswith(ext_bz2):
                basis_fmt = k
                break

        else:
            raise RuntimeError("Unable to determine basis set format of '{}'".format(file_path))
    else:
        basis_fmt = basis_fmt.lower()

    if basis_fmt not in _reader_map:
        raise RuntimeError("Unknown file format to read '{}'".format(basis_fmt))

    # Handle compressed files
    if file_path.endswith('.bz2'):
        with bz2.open(file_path, 'rt', encoding=encoding) as f:
            basis_str = f.read()
    else:
        with open(file_path, 'r', encoding=encoding) as f:
            basis_str = f.read()

    return read_formatted_basis_str(basis_str, basis_fmt, validate, as_component)


def get_reader_formats():
    '''
    Returns the basis set formats that can be read by this library.

    This is returned as an ordered dictionary of key to display name.
    '''

    return {k: v['display'] for k, v in _reader_map.items()}



'''
Conversion of basis sets to Q-Chem format
'''

from .. import lut, manip, sort, printing


def _determine_pure(basis):
    # starts at d shells
    pure = {}
    for eldata in basis['elements'].values():
        if 'electron_shells' not in eldata:
            continue
        for sh in eldata['electron_shells']:
            for shell_am in sh['angular_momentum']:
                harm = '2'  # cartesian
                if 'spherical' in sh['function_type']:
                    harm = '1'

                if shell_am in pure:
                    pure[shell_am] = harm if harm == '1' else pure[shell_am]
                else:
                    pure[shell_am] = harm

    pure_list = sorted(pure.items(), reverse=True)
    pure_list = pure_list[:-2]  # Trim s & p
    return ''.join(x[1] for x in pure_list)


def write_qchem(basis):
    '''Converts a basis set to Q-Chem

    Q-Chem is basically gaussian format, wrapped in $basis/$end

    This also outputs the PURECART variable of the $rem block
    '''

    s = ''
    basis = manip.uncontract_general(basis, True)
    basis = manip.uncontract_spdf(basis, 1, False)
    basis = sort.sort_basis(basis, False)

    # Elements for which we have electron basis
    electron_elements = [k for k, v in basis['elements'].items() if 'electron_shells' in v]

    # Elements for which we have ECP
    ecp_elements = [k for k, v in basis['elements'].items() if 'ecp_potentials' in v]

    purecart = _determine_pure(basis)
    if purecart != '':
        s += "$rem\n"
        if electron_elements:
            s += "    BASIS GEN\n"
        if ecp_elements:
            s += "    ECP GEN\n"
        s += "    PURECART " + _determine_pure(basis) + "\n"
        s += "$end\n\n"

    # Electron Basis
    if electron_elements:
        s += "$basis\n"
        for z in electron_elements:
            data = basis['elements'][z]

            sym = lut.element_sym_from_Z(z, True)
            s += '{}     0\n'.format(sym)

            for shell in data['electron_shells']:
                exponents = shell['exponents']
                coefficients = shell['coefficients']
                ncol = len(coefficients) + 1
                nprim = len(exponents)

                am = shell['angular_momentum']
                amchar = lut.amint_to_char(am, hij=True).upper()

                s += '{}   {}   1.00\n'.format(amchar, nprim)

                point_places = [8 * i + 15 * (i - 1) for i in range(1, ncol + 1)]
                s += printing.write_matrix([exponents, *coefficients], point_places, convert_exp=True)

            s += '****\n'
        s += "$end\n"

    # Write out ECP
    if ecp_elements:
        s += "\n\n$ecp\n"
        for z in ecp_elements:
            data = basis['elements'][z]
            sym = lut.element_sym_from_Z(z).upper()
            max_ecp_am = max([x['angular_momentum'][0] for x in data['ecp_potentials']])
            max_ecp_amchar = lut.amint_to_char([max_ecp_am], hij=True)

            # Sort lowest->highest, then put the highest at the beginning
            ecp_list = sorted(data['ecp_potentials'], key=lambda x: x['angular_momentum'])
            ecp_list.insert(0, ecp_list.pop())

            s += '{}     0\n'.format(sym)
            s += '{}-ECP     {}     {}\n'.format(sym, max_ecp_am, data['ecp_electrons'])

            for pot in ecp_list:
                rexponents = pot['r_exponents']
                gexponents = pot['gaussian_exponents']
                coefficients = pot['coefficients']
                nprim = len(rexponents)

                am = pot['angular_momentum']
                amchar = lut.amint_to_char(am, hij=True)

                if am[0] == max_ecp_am:
                    s += '{} potential\n'.format(amchar)
                else:
                    s += '{}-{} potential\n'.format(amchar, max_ecp_amchar)

                s += '  ' + str(nprim) + '\n'

                point_places = [0, 9, 32]
                s += printing.write_matrix([rexponents, gexponents, *coefficients], point_places, convert_exp=True)
            s += '****\n'
        s += "$end\n"

    return s
